import keras as K
from keras import layers, models, regularizers, backend
from keras.layers import (
    Activation,
    BatchNormalization,
    Conv1D,
    Conv2D,
    Conv3D,
    ConvLSTM1D,
    ConvLSTM2D,
)
from keras.layers.core import Lambda

reg_weight = 1.0e-4


class Concat(layers.Layer):
    def __init__(self, axis=-1, name="Concat", **kwargs):

        self.axis = axis
        super().__init__(name=name)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "axis": self.axis,
                "name": self.name,
            }
        )
        return config

    def call(self, x):

        y = Lambda(lambda x: layers.concatenate([x[0], x[1]], self.axis))(x)

        return y


def LRNet(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=32,
    input_weights=None,
    last_activation="softmax",
):
    """ResNet Version 2 Model builder [b]
    depth of 29 == max pooling of 28 for image patch of 55
    depth of 56 == max pooling of 14 for image patch of 55
    """
    img_input = layers.Input(
        shape=(input_shape[0], None, None, input_shape[3])
    )
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    last_conv_factor = 2 ** (stage_number - 1)
    return_sequences = True
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_3d_lstm_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        return_sequences=return_sequences,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):

            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_3d_lstm_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                return_sequences=return_sequences,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = resnet_3d_lstm_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                return_sequences=return_sequences,
                conv_first=False,
            )
            y = resnet_3d_lstm_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                return_sequences=return_sequences,
                conv_first=False,
            )

            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims

                x = resnet_3d_lstm_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    return_sequences=return_sequences,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])
        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = resnet_3d_lstm_layer(
        inputs=x,
        num_filters=num_filters_out,
        kernel_size=mid_kernel,
        return_sequences=False,
        conv_first=False,
    )

    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, categories:])(x)

    output_cat = (
        Conv2D(
            categories,
            (
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
            name="yolo",
        )
    )(input_cat)
    output_box = (
        Conv2D(
            nboxes * (box_vector),
            (
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
            name="secyolo",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model


def VollNet(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=7,
    mid_kernel=3,
    startfilter=32,
    input_weights=None,
    last_activation="softmax",
):

    img_input = layers.Input(shape=(None, None, None, input_shape[3]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    last_conv_factor = 2 ** (stage_number - 1)
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_3D_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_3D_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = resnet_3D_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = resnet_3D_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_3D_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])
        num_filters_in = num_filters_out
    x = (
        Conv3D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, :, categories:])(x)

    output_cat = (
        Conv3D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
            name="yolo",
        )
    )(input_cat)
    output_box = (
        Conv3D(
            nboxes * (box_vector),
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
            name="secyolo",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model



def DenseVollNet(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    depth=[6, 12, 24, 16],
    start_kernel=7,
    mid_kernel=3,
    startfilter=32,
    stage_number = 3,
    input_weights=None,
    last_activation="softmax",
):

    img_input = layers.Input(shape=(None, None, None, input_shape[3]))
    bn_axis = -1
    num_filters_in = startfilter
    x = densenet_3D_layer( 
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        strides = 2)
    # Start model definition.
    last_conv_factor = 2 ** (stage_number - 1)
    
    x = dense_block(x, depth[0], name="conv2")
    x = transition_block(x, 0.5, name="pool2")
    
    x = dense_block(x, depth[1], name="conv3")
    x = transition_block(x, 0.5, name="pool3")
    
    x = dense_block(x, depth[2], name="conv4")
    

    x = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name="bn")(x)
    x = layers.Activation("relu", name="relu")(x)
    x = (
        Conv3D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, :, categories:])(x)

    output_cat = (
        Conv3D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
            name="yolo",
        )
    )(input_cat)
    output_box = (
        Conv3D(
            nboxes * (box_vector),
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
                round(input_shape[2] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
            name="secyolo",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model



def resnet_lstm_v2(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_weights=None,
    last_activation="softmax",
):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(
        shape=(input_shape[0], input_shape[1], input_shape[2])
    )
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    return_sequences = True
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_lstm_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        return_sequences=return_sequences,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):

            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_lstm_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                return_sequences=return_sequences,
                conv_first=False,
            )
            y = resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_in,
                return_sequences=return_sequences,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_out,
                return_sequences=return_sequences,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims

                x = resnet_lstm_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    return_sequences=return_sequences,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = resnet_lstm_layer(
        inputs=x,
        num_filters=num_filters_out,
        kernel_size=mid_kernel,
        return_sequences=False,
        conv_first=False,
    )
    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, categories:])(x)

    output_cat = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)
    output_box = (
        Conv2D(
            nboxes * (box_vector),
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model


def resnet_v2(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_weights=None,
    last_activation="softmax",
):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(shape=(None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)
    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):

            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = resnet_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = resnet_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims

                x = resnet_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    input_box = Lambda(lambda x: x[:, :, :, categories:])(x)

    output_cat = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)
    output_box = (
        Conv2D(
            nboxes * (box_vector),
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation="sigmoid",
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_box)

    block = Concat(-1)
    outputs = block([output_cat, output_box])

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model


def resnet_1D_regression(
    input_shape,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_weights=None,
):
    """
    # Returns
        model (Model): Keras model instance
    """

    img_input = layers.Input(shape=input_shape)
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_layer_1D(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_layer_1D(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = resnet_layer_1D(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = resnet_layer_1D(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_layer_1D(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling
    x = (
        Conv2D(
            1,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)

    outputs = x

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model


def resnet_v2_class(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_weights=None,
    last_activation="softmax",
):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(shape=(None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = resnet_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = resnet_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out
    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    outputs = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model


def resnet_lstm_v2_class(
    input_shape,
    categories,
    box_vector,
    nboxes=1,
    stage_number=3,
    depth=38,
    start_kernel=3,
    mid_kernel=3,
    startfilter=48,
    input_weights=None,
    last_activation="softmax",
):
    """ResNet Version 2 Model builder [b]
    Stacks of (1 x 1)-(3 x 3)-(1 x 1) BN-ReLU-Conv2D or also known as
    bottleneck layer
    First shortcut connection per layer is 1 x 1 Conv2D.
    Second and onwards shortcut connection is identity.
    At the beginning of each stage, the feature map size is halved (downsampled)
    by a convolutional layer with strides=2, while the number of filter maps is
    doubled. Within each stage, the layers have the same number filters and the
    same filter map sizes.
    Features maps sizes:
    conv1  : 32x32,  16
    stage 0: 32x32,  64
    stage 1: 16x16, 128
    stage 2:  8x8,  256
    # Arguments
        input_shape (tensor): shape of input image tensor
        depth (int): number of core convolutional layers
        num_classes (int): number of classes (CIFAR10 has 10)
    # Returns
        model (Model): Keras model instance
    """

    last_conv_factor = 2 ** (stage_number - 1)
    img_input = layers.Input(shape=(None, None, input_shape[2]))
    if (depth - 2) % 9 != 0:
        raise ValueError("depth should be 9n+2 (eg 56 or 110 in [b])")
    # Start model definition.
    num_filters_in = startfilter
    num_res_blocks = int((depth - 2) / 9)

    # v2 performs Conv2D with BN-ReLU on input before splitting into 2 paths
    x = resnet_lstm_layer(
        inputs=img_input,
        num_filters=num_filters_in,
        kernel_size=start_kernel,
        conv_first=True,
    )

    # Instantiate the stack of residual units
    for stage in range(stage_number):
        for res_block in range(num_res_blocks):
            activation = "relu"
            batch_normalization = True
            strides = 1
            if stage == 0:
                num_filters_out = num_filters_in * 4
                if res_block == 0:  # first layer and first stage
                    activation = None
                    batch_normalization = False
            else:
                num_filters_out = num_filters_in * 2
                if res_block == 0:  # not first layer and not first stage
                    strides = 2  # downsample

            # bottleneck residual unit
            y = resnet_lstm_layer(
                inputs=x,
                num_filters=num_filters_in,
                kernel_size=1,
                strides=strides,
                activation=activation,
                batch_normalization=batch_normalization,
                conv_first=False,
            )
            y = resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_in,
                kernel_size=mid_kernel,
                conv_first=False,
            )
            y = resnet_lstm_layer(
                inputs=y,
                num_filters=num_filters_out,
                kernel_size=1,
                conv_first=False,
            )
            if res_block == 0:
                # linear projection residual shortcut connection to match
                # changed dims
                x = resnet_lstm_layer(
                    inputs=x,
                    num_filters=num_filters_out,
                    kernel_size=1,
                    strides=strides,
                    activation=None,
                    batch_normalization=False,
                )

            x = K.layers.add([x, y])

        num_filters_in = num_filters_out

    # Add classifier on top.
    # v2 has BN-ReLU before Pooling

    x = resnet_lstm_layer(
        inputs=x,
        num_filters=num_filters_out,
        kernel_size=mid_kernel,
        return_sequences=False,
        conv_first=False,
    )
    x = (
        Conv2D(
            categories + nboxes * box_vector,
            kernel_size=mid_kernel,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="same",
        )
    )(x)
    x = BatchNormalization()(x)
    x = Activation("relu")(x)
    input_cat = Lambda(lambda x: x[:, :, :, 0:categories])(x)
    outputs = (
        Conv2D(
            categories,
            (
                round(input_shape[0] / last_conv_factor),
                round(input_shape[1] / last_conv_factor),
            ),
            activation=last_activation,
            kernel_regularizer=regularizers.l2(reg_weight),
            padding="valid",
        )
    )(input_cat)

    inputs = img_input

    # Create model.
    model = models.Model(inputs, outputs)

    if input_weights is not None:

        model.load_weights(input_weights, by_name=True)

    return model


def resnet_3d_lstm_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    return_sequences=True,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):

    conv_lstm_3d = ConvLSTM2D(
        filters=num_filters,
        kernel_size=(kernel_size, kernel_size),
        activation=activation,
        strides=strides,
        data_format="channels_last",
        return_sequences=return_sequences,
        padding="same",
    )
    x = inputs
    if conv_first:
        x = conv_lstm_3d(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv_lstm_3d(x)
    return x


def resnet_lstm_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    return_sequences=True,
    batch_normalization=True,
    conv_first=True,
):

    conv_lstm = ConvLSTM1D(
        filters=num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        return_sequences=return_sequences,
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv_lstm(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv_lstm(x)
    return x


def resnet_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):
    """2D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv2D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x


def resnet_layer_1D(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):
    """1D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv1D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x


def dense_block(x, blocks, name):
    """A dense block.
    Args:
      x: input tensor.
      blocks: integer, the number of building blocks.
      name: string, block label.
    Returns:
      Output tensor for the block.
    """
    for i in range(blocks):
        x = dense_conv_block(x, 32, name=name + "_block" + str(i + 1))
    return x


def transition_block(x, reduction, name):
    """A transition block.
    Args:
      x: input tensor.
      reduction: float, compression rate at transition layers.
      name: string, block label.
    Returns:
      output tensor for the block.
    """
    bn_axis = -1 
    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=name + "_bn"
    )(x)
    x = layers.Activation("relu", name=name + "_relu")(x)
    x = layers.Conv3D(
        int(backend.int_shape(x)[bn_axis] * reduction),
        1,
        use_bias=False, padding = 'same',
        name=name + "_conv"
    )(x)
    x = layers.AveragePooling3D(2, strides=2, name=name + "_pool", padding = 'same')(x)
    return x


def dense_conv_block(x, growth_rate, name):
    """A building block for a dense block.
    Args:
      x: input tensor.
      growth_rate: float, growth rate at dense layers.
      name: string, block label.
    Returns:
      Output tensor for the block.
    """
    bn_axis = -1 
    x1 = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=name + "_0_bn"
    )(x)
    x1 = layers.Activation("relu", name=name + "_0_relu")(x1)
    x1 = layers.Conv3D(
        4 * growth_rate, 1, use_bias=False, name=name + "_1_conv", padding = 'same'
    )(x1)
    x1 = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=name + "_1_bn"
    )(x1)
    x1 = layers.Activation("relu", name=name + "_1_relu")(x1)
    x1 = layers.Conv3D(
        growth_rate, 3, use_bias=False, name=name + "_2_conv",padding = 'same'
    )(x1)
    x = layers.Concatenate(axis=bn_axis, name=name + "_concat")([x, x1])
    return x



def densenet_3D_layer(
    inputs,
    num_filters = 64,
    kernel_size = 3,
    strides = 1,
    activation='relu',
   
):
    
    x = inputs
    x = layers.Conv3D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )(inputs)
    x = layers.BatchNormalization(
        axis= -1, epsilon=1.001e-5, name="conv1/bn"
    )(x)
    x = layers.Activation(activation, name="conv1activation")(x)
    
    return x


def dense_block(x, blocks, name):
    """A dense block.
    Args:
      x: input tensor.
      blocks: integer, the number of building blocks.
      name: string, block label.
    Returns:
      Output tensor for the block.
    """
    for i in range(blocks):
        x = dense_conv_block(x, 32, name=name + "_block" + str(i + 1))
    return x


def transition_block(x, reduction, name):
    """A transition block.
    Args:
      x: input tensor.
      reduction: float, compression rate at transition layers.
      name: string, block label.
    Returns:
      output tensor for the block.
    """
    bn_axis = -1 
    x = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=name + "_bn"
    )(x)
    x = layers.Activation("relu", name=name + "_relu")(x)
    x = layers.Conv3D(
        int(backend.int_shape(x)[bn_axis] * reduction),
        1,
        use_bias=False, padding = 'same',
        name=name + "_conv"
    )(x)
    x = layers.AveragePooling3D(2, strides=2, name=name + "_pool", padding = 'same')(x)
    return x


def dense_conv_block(x, growth_rate, name):
    """A building block for a dense block.
    Args:
      x: input tensor.
      growth_rate: float, growth rate at dense layers.
      name: string, block label.
    Returns:
      Output tensor for the block.
    """
    bn_axis = -1 
    x1 = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=name + "_0_bn"
    )(x)
    x1 = layers.Activation("relu", name=name + "_0_relu")(x1)
    x1 = layers.Conv3D(
        4 * growth_rate, 1, use_bias=False, name=name + "_1_conv", padding = 'same'
    )(x1)
    x1 = layers.BatchNormalization(
        axis=bn_axis, epsilon=1.001e-5, name=name + "_1_bn"
    )(x1)
    x1 = layers.Activation("relu", name=name + "_1_relu")(x1)
    x1 = layers.Conv3D(
        growth_rate, 3, use_bias=False, name=name + "_2_conv",padding = 'same'
    )(x1)
    x = layers.Concatenate(axis=bn_axis, name=name + "_concat")([x, x1])
    return x



def densenet_3D_layer(
    inputs,
    num_filters = 64,
    kernel_size = 3,
    strides = 1,
    activation='relu',
   
):
    
    x = inputs
    x = layers.Conv3D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )(inputs)
    x = layers.BatchNormalization(
        axis= -1, epsilon=1.001e-5, name="conv1/bn"
    )(x)
    x = layers.Activation(activation, name="conv1activation")(x)
    
    return x



def resnet_3D_layer(
    inputs,
    num_filters=64,
    kernel_size=3,
    strides=1,
    activation="relu",
    batch_normalization=True,
    conv_first=True,
):
    """3D Convolution-Batch Normalization-Activation stack builder
    # Arguments
        inputs (tensor): input tensor from input image or previous layer
        num_filters (int): Conv2D number of filters
        kernel_size (int): Conv2D square kernel dimensions
        strides (int): Conv2D square stride dimensions
        activation (string): activation name
        batch_normalization (bool): whether to include batch normalization
        conv_first (bool): conv-bn-activation (True) or
            bn-activation-conv (False)
    # Returns
        x (tensor): tensor as input to the next layer
    """
    conv = Conv3D(
        num_filters,
        kernel_size=kernel_size,
        strides=strides,
        padding="same",
        kernel_initializer="he_normal",
        kernel_regularizer=regularizers.l2(1e-4),
    )

    x = inputs
    if conv_first:
        x = conv(x)
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
    else:
        if batch_normalization:
            x = BatchNormalization()(x)
        if activation is not None:
            x = Activation(activation)(x)
        x = conv(x)
    return x



#!/usr/bin/env python2
"""
Created on Tue Mar 30 16:05:34 2021
@author: vkapoor
"""

import numpy as np
import tensorflow as tf
from keras import backend as K

lambdaobject = 1
lambdanoobject = 1
lambdacoord = 1
lambdaclass = 1
lambdaangle = 1


def get_event_grid(grid_h, grid_w, grid_t, boxes):

    event_grid = np.array(
        [
            [[float(y), float(x), float(t)]] * boxes
            for y in range(grid_h)
            for x in range(grid_w)
            for t in range(grid_t)
        ]
    )

    return event_grid


def extract_ground_event_truth(
    y_true, categories, grid_h, grid_w, grid_t, nboxes, box_vector
):

    true_box_class = y_true[..., 0:categories]

    true_nboxes = K.reshape(
        y_true[..., categories:],
        (-1, grid_h * grid_w * grid_t, nboxes, box_vector),
    )

    true_box_xyt = true_nboxes[..., 0:3]

    true_box_wh = true_nboxes[..., 3:5]

    true_box_conf = true_nboxes[..., -1]

    return true_box_class, true_box_xyt, true_box_wh, true_box_conf


def extract_ground_event_pred(
    y_pred, categories, grid_h, grid_w, grid_t, event_grid, nboxes, box_vector
):

    pred_box_class = y_pred[..., 0:categories]

    pred_nboxes = K.reshape(
        y_pred[..., categories:],
        (-1, grid_h * grid_w * grid_t, nboxes, box_vector),
    )

    pred_box_xyt = pred_nboxes[..., 0:3] + event_grid

    pred_box_wh = pred_nboxes[..., 3:5]

    pred_box_conf = pred_nboxes[..., -1]

    return pred_box_class, pred_box_xyt, pred_box_wh, pred_box_conf


def extract_ground_event_volume_truth(
    y_true, categories, grid_h, grid_w, grid_d, nboxes, box_vector
):

    true_box_class = y_true[..., 0:categories]

    true_nboxes = K.reshape(
        y_true[..., categories:],
        (-1, grid_h * grid_w * grid_d, nboxes, box_vector),
    )

    true_box_xyz = true_nboxes[..., 0:3]

    true_box_whd = true_nboxes[..., 4:7]

    true_box_conf = true_nboxes[..., -1]

    return true_box_class, true_box_xyz, true_box_whd, true_box_conf


def extract_ground_event_volume_pred(
    y_pred, categories, grid_h, grid_w, grid_d, event_grid, nboxes, box_vector
):

    pred_box_class = y_pred[..., 0:categories]

    pred_nboxes = K.reshape(
        y_pred[..., categories:],
        (-1, grid_h * grid_w * grid_d, nboxes, box_vector),
    )

    pred_box_xyz = pred_nboxes[..., 0:3] + event_grid

    pred_box_whd = pred_nboxes[..., 4:7]

    pred_box_conf = pred_nboxes[..., -1]

    return pred_box_class, pred_box_xyz, pred_box_whd, pred_box_conf


def get_cell_grid(grid_h, grid_w, boxes):

    cell_grid = np.array(
        [
            [[float(x), float(y)]] * boxes
            for y in range(grid_h)
            for x in range(grid_w)
        ]
    )

    return cell_grid


def extract_ground_cell_pred(
    y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector
):

    pred_box_class = y_pred[..., 0:categories]

    pred_nboxes = K.reshape(
        y_pred[..., categories:], (-1, grid_h * grid_w, nboxes, box_vector)
    )

    pred_box_xy = pred_nboxes[..., 0:2] + (cell_grid)

    pred_box_wh = pred_nboxes[..., -3:-1]

    pred_box_conf = pred_nboxes[..., -1]

    return pred_box_class, pred_box_xy, pred_box_wh, pred_box_conf


def extract_ground_cell_truth(
    y_truth, categories, grid_h, grid_w, nboxes, box_vector
):

    true_box_class = y_truth[..., 0:categories]

    true_nboxes = K.reshape(
        y_truth[..., categories:], (-1, grid_h * grid_w, nboxes, box_vector)
    )

    true_box_xy = true_nboxes[..., 0:2]

    true_box_wh = true_nboxes[..., -3:-1]

    true_box_conf = true_nboxes[..., -1]

    return true_box_class, true_box_xy, true_box_wh, true_box_conf


def extract_ground_cell_pred_class(y_pred, categories):

    pred_box_class = y_pred[..., 0:categories]

    return pred_box_class


def extract_ground_cell_pred_segfree(
    y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector
):

    pred_box_class = y_pred[..., 0:categories]

    pred_nboxes = K.reshape(
        y_pred[..., categories:], (-1, grid_h * grid_w, nboxes, box_vector)
    )

    pred_box_xy = pred_nboxes[..., 0:2] + cell_grid

    return pred_box_class, pred_box_xy


def extract_ground_cell_truth_foc(y_truth, categories):

    true_box_class = y_truth[..., 0:categories]

    return true_box_class


def extract_ground_cell_pred_foc(y_pred, categories):

    pred_box_class = y_pred[..., 0:categories]

    return pred_box_class


def extract_ground_cell_truth_segfree(
    y_truth, categories, grid_h, grid_w, nboxes, box_vector
):

    true_box_class = y_truth[..., 0:categories]

    true_nboxes = K.reshape(
        y_truth[..., categories:], (-1, grid_h * grid_w, nboxes, box_vector)
    )

    true_box_xy = true_nboxes[..., 0:2]

    return true_box_class, true_box_xy


def extract_ground_cell_truth_class(y_truth, categories):

    true_box_class = y_truth[..., 0:categories]

    return true_box_class


def compute_conf_loss(
    pred_box_wh,
    true_box_wh,
    pred_box_xy,
    true_box_xy,
    true_box_conf,
    pred_box_conf,
):

    # compute the intersection of all boxes at once (the IOU)
    intersect_wh = K.maximum(
        K.zeros_like(pred_box_wh),
        (pred_box_wh + true_box_wh) / 2
        - K.square(pred_box_xy[..., 0:1] - true_box_xy[..., 0:1]),
    )
    intersect_area = intersect_wh[..., 0] * intersect_wh[..., 1]
    true_area = true_box_wh[..., 0] * true_box_wh[..., 1]
    pred_area = pred_box_wh[..., 0] * pred_box_wh[..., 1]
    union_area = pred_area + true_area - intersect_area
    iou = tf.truediv(intersect_area, union_area)

    best_ious = K.max(iou, axis=-1)
    loss_conf = K.sum(
        K.square(true_box_conf * best_ious - pred_box_conf), axis=-1
    )

    loss_conf = loss_conf * lambdaobject

    return loss_conf


def compute_conf_loss_volume(
    pred_box_whd,
    true_box_whd,
    pred_box_xyz,
    true_box_xyz,
    true_box_conf,
    pred_box_conf,
):

    # compute the intersection of all boxes at once (the IOU)
    intersect_whd = K.maximum(
        K.zeros_like(pred_box_whd),
        (pred_box_whd + true_box_whd) / 2
        - K.square(pred_box_xyz[..., 0:3] - true_box_xyz[..., 0:3]),
    )
    intersect_area = (
        intersect_whd[..., 0] * intersect_whd[..., 1] * intersect_whd[..., 2]
    )
    true_area = (
        true_box_whd[..., 0] * true_box_whd[..., 1] * true_box_whd[..., 2]
    )
    pred_area = (
        pred_box_whd[..., 0] * pred_box_whd[..., 1] * pred_box_whd[..., 2]
    )
    union_area = pred_area + true_area - intersect_area
    iou = tf.truediv(intersect_area, union_area)

    best_ious = K.max(iou, axis=-1)
    loss_conf = K.sum(
        K.square(true_box_conf * best_ious - pred_box_conf), axis=-1
    )

    loss_conf = loss_conf * lambdaobject

    return loss_conf


def compute_conf_loss_static(
    pred_box_wh,
    true_box_wh,
    pred_box_xy,
    true_box_xy,
    true_box_conf,
    pred_box_conf,
):

    # compute the intersection of all boxes at once (the IOU)
    intersect_wh = K.maximum(
        K.zeros_like(pred_box_wh),
        (pred_box_wh + true_box_wh) / 2 - K.square(pred_box_xy - true_box_xy),
    )
    intersect_area = intersect_wh[..., 0] * intersect_wh[..., 1]
    true_area = true_box_wh[..., 0] * true_box_wh[..., 1]
    pred_area = pred_box_wh[..., 0] * pred_box_wh[..., 1]
    union_area = pred_area + true_area - intersect_area
    iou = intersect_area / union_area

    loss_conf = K.sum(K.square(true_box_conf * iou - pred_box_conf), axis=-1)

    loss_conf = loss_conf * lambdaobject

    return loss_conf


def calc_loss_xywh(true_box_xy, pred_box_xy, true_box_wh, pred_box_wh):

    loss_xy = K.sum(
        K.sum(K.square(true_box_xy - pred_box_xy), axis=-1), axis=-1
    )
    loss_wh = K.sum(
        K.sum(K.square(K.sqrt(true_box_wh) - K.sqrt(pred_box_wh)), axis=-1),
        axis=-1,
    )
    loss_xywh = loss_xy + loss_wh
    loss_xywh = lambdacoord * loss_xywh
    return loss_xywh


def calc_loss_xyzwhd(true_box_xyz, pred_box_xyz, true_box_whd, pred_box_whd):

    loss_xyz = K.sum(
        K.sum(K.square(true_box_xyz - pred_box_xyz), axis=-1), axis=-1
    )
    loss_whd = K.sum(
        K.sum(K.square(K.sqrt(true_box_whd) - K.sqrt(pred_box_whd)), axis=-1),
        axis=-1,
    )
    loss_xyzwhd = loss_xyz + loss_whd
    loss_xyzwhd = lambdacoord * loss_xyzwhd
    return loss_xyzwhd


def calc_loss_xy(true_box_xy, pred_box_xy):

    loss_xy = K.sum(
        K.sum(K.square(true_box_xy - pred_box_xy), axis=-1), axis=-1
    )
    loss_xy = lambdacoord * loss_xy
    return loss_xy


def calc_loss_angle(true_box_angle, pred_box_angle):

    loss_angle = K.sum(
        K.sum(K.square(true_box_angle - pred_box_angle), axis=-1), axis=-1
    )
    loss_angle = lambdaangle * loss_angle
    return loss_angle


def calc_loss_class(true_box_class, pred_box_class, entropy):

    if entropy == "binary":
        loss_class = K.mean(
            K.binary_crossentropy(true_box_class, pred_box_class), axis=-1
        )
    if entropy == "notbinary":
        loss_class = K.mean(
            K.categorical_crossentropy(true_box_class, pred_box_class), axis=-1
        )

    loss_class = loss_class * lambdaclass

    return loss_class


def dynamic_yolo_loss(
    categories, grid_h, grid_w, grid_t, nboxes, box_vector, entropy
):
    def loss(y_true, y_pred):
        event_grid = get_event_grid(grid_h, grid_w, grid_t, nboxes)
        (
            true_box_class,
            true_box_xyt,
            true_box_wh,
            true_box_conf,
        ) = extract_ground_event_truth(
            y_true, categories, grid_h, grid_w, grid_t, nboxes, box_vector
        )
        (
            pred_box_class,
            pred_box_xyt,
            pred_box_wh,
            pred_box_conf,
        ) = extract_ground_event_pred(
            y_pred,
            categories,
            grid_h,
            grid_w,
            grid_t,
            event_grid,
            nboxes,
            box_vector,
        )

        loss_xywht = calc_loss_xywh(
            true_box_xyt, pred_box_xyt, true_box_wh, pred_box_wh
        )

        loss_class = calc_loss_class(true_box_class, pred_box_class, entropy)

        loss_conf = compute_conf_loss(
            pred_box_wh,
            true_box_wh,
            pred_box_xyt,
            true_box_xyt,
            true_box_conf,
            pred_box_conf,
        )
        combinedloss = loss_xywht + loss_conf + loss_class

        return combinedloss

    return loss


def volume_yolo_loss(
    categories, grid_h, grid_w, grid_d, nboxes, box_vector, entropy
):
    def loss(y_true, y_pred):
        event_grid = get_event_grid(grid_h, grid_w, grid_d, nboxes)
        (
            true_box_class,
            true_box_xyz,
            true_box_whd,
            true_box_conf,
        ) = extract_ground_event_volume_truth(
            y_true, categories, grid_h, grid_w, grid_d, nboxes, box_vector
        )
        (
            pred_box_class,
            pred_box_xyz,
            pred_box_whd,
            pred_box_conf,
        ) = extract_ground_event_volume_pred(
            y_pred,
            categories,
            grid_h,
            grid_w,
            grid_d,
            event_grid,
            nboxes,
            box_vector,
        )

        loss_xywht = calc_loss_xyzwhd(
            true_box_xyz, pred_box_xyz, true_box_whd, pred_box_whd
        )

        loss_class = calc_loss_class(true_box_class, pred_box_class, entropy)

        loss_conf = compute_conf_loss_volume(
            pred_box_whd,
            true_box_whd,
            pred_box_xyz,
            true_box_xyz,
            true_box_conf,
            pred_box_conf,
        )
        # Adding it all up
        combinedloss = loss_xywht + loss_conf + loss_class

        return combinedloss

    return loss


def static_yolo_loss(categories, grid_h, grid_w, nboxes, box_vector, entropy):
    def loss(y_true, y_pred):

        cell_grid = get_cell_grid(grid_h, grid_w, nboxes)
        (
            true_box_class,
            true_box_xy,
            true_box_wh,
            true_box_conf,
        ) = extract_ground_cell_truth(
            y_true, categories, grid_h, grid_w, nboxes, box_vector
        )
        (
            pred_box_class,
            pred_box_xy,
            pred_box_wh,
            pred_box_conf,
        ) = extract_ground_cell_pred(
            y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector
        )

        loss_xywh = calc_loss_xywh(
            true_box_conf, true_box_xy, pred_box_xy, true_box_wh, pred_box_wh
        )

        loss_class = calc_loss_class(true_box_class, pred_box_class, entropy)

        loss_conf = compute_conf_loss_static(
            pred_box_wh,
            true_box_wh,
            pred_box_xy,
            true_box_xy,
            true_box_conf,
            pred_box_conf,
        )

        combinedloss = loss_xywh + loss_conf + loss_class

        return combinedloss

    return loss


def static_yolo_loss_segfree(
    categories, grid_h, grid_w, nboxes, box_vector, entropy
):
    def loss(y_true, y_pred):

        cell_grid = get_cell_grid(grid_h, grid_w, nboxes)
        true_box_class, true_box_xy = extract_ground_cell_truth_segfree(
            y_true, categories, grid_h, grid_w, nboxes, box_vector
        )
        pred_box_class, pred_box_xy = extract_ground_cell_pred_segfree(
            y_pred, categories, grid_h, grid_w, cell_grid, nboxes, box_vector
        )
        loss_xy = calc_loss_xy(true_box_xy, pred_box_xy)

        loss_class = calc_loss_class(true_box_class, pred_box_class, entropy)

        combinedloss = loss_xy + loss_class

        return combinedloss

    return loss


def class_yolo_loss(categories, entropy):
    def loss(y_true, y_pred):

        true_box_class = extract_ground_cell_truth_class(y_true, categories)
        pred_box_class = extract_ground_cell_pred_class(y_pred, categories)
        loss_class = calc_loss_class(true_box_class, pred_box_class, entropy)
        combinedloss = loss_class

        return combinedloss

    return loss



import csv
import glob
import json
import os
from pathlib import Path

import numpy as np
import pandas as pd
from PIL import Image
from skimage import measure
from sklearn.model_selection import train_test_split
from tifffile import imread, imwrite
from tqdm import tqdm

from .utils import normalizeFloatZeroOne


def SegFreeMovieLabelDataSet(
    image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    normPatch=False,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:

        name = os.path.basename(os.path.splitext(fname)[0])

        for csvfname in filesCsv:
            count = 0
            Csvname = os.path.basename(os.path.splitext(csvfname)[0])

            for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                classfound = Csvname == csv_name_diff + event_name + name
                if classfound:
                    print(Csvname)
                    image = imread(fname).astype(dtype)
                    if not normPatch and normalizeimage:
                        image = normalizeFloatZeroOne(
                            image.astype(dtype), 1, 99.8, dtype=dtype
                        )
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]][1:]
                    y = dataset[dataset.keys()[1]][1:]
                    x = dataset[dataset.keys()[2]][1:]

                    # Categories + XYHW + Confidence
                    for (key, t) in time.items():

                        SimpleMovieMaker(
                            t,
                            y[key],
                            x[key],
                            image,
                            crop_size,
                            gridx,
                            gridy,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                            normPatch,
                            tshift,
                        )
                        count = count + 1


def SegFreeMovieLabelDataSet4D(
    image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    normPatch=False,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:

        name = os.path.basename(os.path.splitext(fname)[0])

        for csvfname in filesCsv:
            count = 0
            Csvname = os.path.basename(os.path.splitext(csvfname)[0])

            for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                classfound = Csvname == csv_name_diff + event_name + name
                if classfound:
                    print(Csvname)
                    image = imread(fname).astype(dtype)
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]][1:]
                    z = dataset[dataset.keys()[1]][1:]
                    y = dataset[dataset.keys()[2]][1:]
                    x = dataset[dataset.keys()[3]][1:]

                    # Categories + XYHW + Confidence
                    for (key, t) in time.items():

                        SimpleMovieMaker4D(
                            normalizeimage,
                            t,
                            z[key],
                            y[key],
                            x[key],
                            image,
                            crop_size,
                            gridx,
                            gridy,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                            normPatch,
                            tshift,
                            dtype,
                        )
                        count = count + 1


def SimpleMovieMaker(
    time,
    y,
    x,
    image,
    crop_size,
    gridx,
    gridy,
    total_categories,
    trainlabel,
    name,
    save_dir,
    normPatch,
    tshift,
    dtype,
):

    sizex, sizey, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if time > 0:
        for shift in AllShifts:

            newname = name + "shift" + str(shift)
            Event_data = []

            Label = np.zeros([total_categories + 6])

            newcenter = (y - shift[1], x - shift[0])
            if (
                x + shift[0] > sizex / 2
                and y + shift[1] > sizey / 2
                and x + shift[0] + int(imagesizex / 2) < image.shape[2]
                and y + shift[1] + int(imagesizey / 2) < image.shape[1]
                and time > size_tminus
                and time + size_tplus + 1 < image.shape[0]
            ):
                crop_xminus = x - int(imagesizex / 2)
                crop_xplus = x + int(imagesizex / 2)
                crop_yminus = y - int(imagesizey / 2)
                crop_yplus = y + int(imagesizey / 2)
                # Cut off the region for training movie creation
                region = (
                    slice(int(time - size_tminus), int(time + size_tplus + 1)),
                    slice(
                        int(crop_yminus) + shift[1], int(crop_yplus) + shift[1]
                    ),
                    slice(
                        int(crop_xminus) + shift[0], int(crop_xplus) + shift[0]
                    ),
                )
                # Define the movie region volume that was cut
                crop_image = image[region]
                if normPatch:
                    crop_image = normalizeFloatZeroOne(
                        crop_image, 1, 99.8, dtype=dtype
                    )
                seglocationx = newcenter[1] - crop_xminus
                seglocationy = newcenter[0] - crop_yminus
                Label[total_categories] = seglocationx / sizex
                Label[total_categories + 1] = seglocationy / sizey
                Label[total_categories + 2] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                Label[total_categories + 3] = 1
                Label[total_categories + 4] = 1
                Label[total_categories + 5] = 1

                # Write the image as 32 bit tif file
                if (
                    crop_image.shape[0] == size_tplus + size_tminus + 1
                    and crop_image.shape[1] == imagesizey
                    and crop_image.shape[2] == imagesizex
                ):

                    imwrite(
                        (save_dir + "/" + newname + ".tif"),
                        crop_image.astype("float32"),
                    )
                    Event_data.append([Label[i] for i in range(0, len(Label))])
                    if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                        os.remove(save_dir + "/" + (newname) + ".csv")
                    writer = csv.writer(
                        open(save_dir + "/" + (newname) + ".csv", "a")
                    )
                    writer.writerows(Event_data)


def SimpleMovieMaker4D(
    normalizeimage,
    time,
    z,
    y,
    x,
    image,
    crop_size,
    gridx,
    gridy,
    total_categories,
    trainlabel,
    name,
    save_dir,
    normPatch,
    tshift,
    dtype,
):

    sizex, sizey, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift

    image = image[:, z, :, :]
    if normalizeimage:
        image = normalizeFloatZeroOne(
            image.astype(dtype), 1, 99.8, dtype=dtype
        )

    if time > 0:

        for shift in AllShifts:

            newname = name + "shift" + str(shift)
            Event_data = []

            Label = np.zeros([total_categories + 6])
            Label[trainlabel] = 1

            newcenter = (y - shift[1], x - shift[0])
            if (
                x + shift[0] > sizex / 2
                and y + shift[1] > sizey / 2
                and x + shift[0] + int(imagesizex / 2) < image.shape[2]
                and y + shift[1] + int(imagesizey / 2) < image.shape[1]
                and time > size_tminus
                and time + size_tplus + 1 < image.shape[0]
            ):
                crop_xminus = x - int(imagesizex / 2)
                crop_xplus = x + int(imagesizex / 2)
                crop_yminus = y - int(imagesizey / 2)
                crop_yplus = y + int(imagesizey / 2)
                # Cut off the region for training movie creation
                region = (
                    slice(int(time - size_tminus), int(time + size_tplus + 1)),
                    slice(
                        int(crop_yminus) + shift[1], int(crop_yplus) + shift[1]
                    ),
                    slice(
                        int(crop_xminus) + shift[0], int(crop_xplus) + shift[0]
                    ),
                )
                # Define the movie region volume that was cut
                crop_image = image[region]
                if normPatch:
                    crop_image = normalizeFloatZeroOne(
                        crop_image, 1, 99.8, dtype=dtype
                    )

                seglocationx = newcenter[1] - crop_xminus
                seglocationy = newcenter[0] - crop_yminus
                Label[total_categories] = seglocationx / sizex
                Label[total_categories + 1] = seglocationy / sizey
                Label[total_categories + 2] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                Label[total_categories + 3] = 1
                Label[total_categories + 4] = 1
                Label[total_categories + 5] = 1

                # Write the image as 32 bit tif file
                if (
                    crop_image.shape[0] == size_tplus + size_tminus + 1
                    and crop_image.shape[1] == imagesizey
                    and crop_image.shape[2] == imagesizex
                ):

                    imwrite(
                        (save_dir + "/" + newname + ".tif"),
                        crop_image.astype("float32"),
                    )
                    Event_data.append([Label[i] for i in range(0, len(Label))])
                    if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                        os.remove(save_dir + "/" + (newname) + ".csv")
                    writer = csv.writer(
                        open(save_dir + "/" + (newname) + ".csv", "a")
                    )
                    writer.writerows(Event_data)


def loadResizeImgs(im, size):

    w, h = im.size
    if w < h:
        im = im.crop((0, (h - w) / 2, w, (h + w) / 2))
    elif w > h:
        im = im.crop(((w - h + 1) / 2, 0, (w + h) / 2, h))

    return np.array(im.resize(size, Image.BILINEAR))


def Folder_to_oneat(
    dir, trainlabel, trainname, total_categories, size, save_dir
):

    Label = np.zeros([total_categories])

    count = 0

    files = sorted(glob.glob(dir + "/" + "*.png"))
    for i in tqdm(range(len(files))):
        file = files[i]
        try:
            Event_data = []

            img = Image.open(file)
            img = loadResizeImgs(img, size)
            Name = str(trainname) + os.path.basename(os.path.splitext(file)[0])
            image = normalizeFloatZeroOne(img.astype("uint16"), 1, 99.8)
            Label[trainlabel] = 1
            imwrite(
                (save_dir + "/" + Name + str(count) + ".tif"),
                image.astype("uint16"),
            )
            Event_data.append([Label[i] for i in range(0, len(Label))])
            if os.path.exists(save_dir + "/" + Name + str(count) + ".csv"):
                os.remove(save_dir + "/" + Name + str(count) + ".csv")
            writer = csv.writer(
                open(save_dir + "/" + Name + str(count) + ".csv", "a")
            )
            writer.writerows(Event_data)
            count = count + 1
        except Exception as e:
            print("[WW] ", str(e))
            continue


def Midog_to_oneat(
    midog_folder,
    annotation_file,
    event_type_name_label,
    all_ids,
    crop_size,
    save_dir,
):

    rows = []
    annotations = {}
    id_to_tumortype = {
        id: list(k for k in all_ids if id in all_ids[k])[0]
        for id in range(1, 406)
    }
    with open(annotation_file) as f:
        data = json.load(f)

        # categories = {cat["id"]: cat["name"] for cat in data["categories"]}
        total_categories = len(event_type_name_label.keys())
        for row in data["images"]:
            file_name = row["file_name"]
            image_id = row["id"]
            width = row["width"]
            height = row["height"]
            tumortype = id_to_tumortype[image_id]

            for annotation in [
                anno
                for anno in data["annotations"]
                if anno["image_id"] == image_id
            ]:
                box = annotation["bbox"]
                cat = annotation["category_id"]

                rows.append([image_id, width, height, box, cat, tumortype])
            annotations[file_name] = rows

    count = 0
    for tumortype, ids in zip(list(all_ids.keys()), list(all_ids.values())):

        for image_id in ids:

            file_path = midog_folder + "/" + f"{image_id:03d}.tiff"
            Name = os.path.basename(os.path.splitext(file_path)[0])

            img = imread(file_path)
            image = normalizeFloatZeroOne(img.astype("uint16"), 1, 99.8)
            image_annotation_array = annotations[Name + ".tiff"]
            image_id += 1

    for image_annotation in image_annotation_array:

        Label = np.zeros([total_categories + 5])
        Event_data = []
        (
            image_id,
            image_width,
            image_height,
            box,
            cat,
            tumortype,
        ) = image_annotation
        Name = str(image_id)
        x0, y0, x1, y1 = box
        height = y1 - y0
        width = x1 - x0
        x = (x0 + x1) // 2
        y = (y0 + y1) // 2
        # if cat == 1 then it is mitosis if cat == 2 it is hard negative
        if cat == 2:
            trainlabel = (
                event_type_name_label[tumortype] + total_categories // 2
            )
        if cat == 1:
            trainlabel = event_type_name_label[tumortype]
        ImagesizeX, ImagesizeY = crop_size
        crop_Xminus = x - int(ImagesizeX / 2)
        crop_Xplus = x + int(ImagesizeX / 2)
        crop_Yminus = y - int(ImagesizeY / 2)
        crop_Yplus = y + int(ImagesizeY / 2)
        region = (
            slice(int(crop_Yminus), int(crop_Yplus)),
            slice(int(crop_Xminus), int(crop_Xplus)),
        )

        crop_image = image[region]

        Label[trainlabel] = 1
        Label[total_categories] = 0.5
        Label[total_categories + 1] = 0.5
        Label[total_categories + 2] = width / ImagesizeX
        Label[total_categories + 3] = height / ImagesizeY
        Label[total_categories + 4] = 1

        count = count + 1
        if (
            crop_image.shape[0] == ImagesizeY
            and crop_image.shape[1] == ImagesizeX
        ):
            imwrite(
                (save_dir + "/" + Name + str(count) + ".tif"),
                crop_image.astype("float32"),
            )
            Event_data.append([Label[i] for i in range(0, len(Label))])
            if os.path.exists(save_dir + "/" + Name + str(count) + ".csv"):
                os.remove(save_dir + "/" + Name + str(count) + ".csv")
            writer = csv.writer(
                open(save_dir + "/" + Name + str(count) + ".csv", "a")
            )
            writer.writerows(Event_data)


def Midog_to_oneat_simple(
    midog_folder,
    annotation_file,
    event_type_name_label,
    all_ids,
    crop_size,
    save_dir,
):

    rows = []
    annotations = {}
    id_to_tumortype = {
        id: list(k for k in all_ids if id in all_ids[k])[0]
        for id in range(1, 406)
    }
    with open(annotation_file) as f:
        data = json.load(f)

        total_categories = len(event_type_name_label.keys())
        for row in data["images"]:
            file_name = row["file_name"]
            image_id = row["id"]
            width = row["width"]
            height = row["height"]
            tumortype = id_to_tumortype[image_id]

            for annotation in [
                anno
                for anno in data["annotations"]
                if anno["image_id"] == image_id
            ]:
                box = annotation["bbox"]
                cat = annotation["category_id"]

                rows.append([image_id, width, height, box, cat, tumortype])
            annotations[file_name] = rows

    count = 0
    for tumortype, ids in zip(list(all_ids.keys()), list(all_ids.values())):

        for image_id in ids:

            file_path = midog_folder + "/" + f"{image_id:03d}.tiff"
            Name = os.path.basename(os.path.splitext(file_path)[0])

            img = imread(file_path)
            image = normalizeFloatZeroOne(img.astype("uint16"), 1, 99.8)
            image_annotation_array = annotations[Name + ".tiff"]
            image_id += 1

    for image_annotation in image_annotation_array:

        Label = np.zeros([total_categories])
        Event_data = []
        (
            image_id,
            image_width,
            image_height,
            box,
            cat,
            tumortype,
        ) = image_annotation
        Name = str(image_id)
        x0, y0, x1, y1 = box
        height = y1 - y0
        width = x1 - x0
        x = (x0 + x1) // 2
        y = (y0 + y1) // 2
        # if cat == 1 then it is mitosis if cat == 2 it is hard negative
        if cat == 2:
            trainlabel = (
                event_type_name_label[tumortype] + total_categories // 2
            )
        if cat == 1:
            trainlabel = event_type_name_label[tumortype]
        ImagesizeX, ImagesizeY = crop_size
        crop_Xminus = x - int(ImagesizeX / 2)
        crop_Xplus = x + int(ImagesizeX / 2)
        crop_Yminus = y - int(ImagesizeY / 2)
        crop_Yplus = y + int(ImagesizeY / 2)
        region = (
            slice(int(crop_Yminus), int(crop_Yplus)),
            slice(int(crop_Xminus), int(crop_Xplus)),
        )

        crop_image = image[region]

        Label[trainlabel] = 1

        count = count + 1
        if (
            crop_image.shape[0] == ImagesizeY
            and crop_image.shape[1] == ImagesizeX
        ):
            imwrite(
                (save_dir + "/" + Name + str(count) + ".tif"),
                crop_image.astype("float32"),
            )
            Event_data.append([Label[i] for i in range(0, len(Label))])
            if os.path.exists(save_dir + "/" + Name + str(count) + ".csv"):
                os.remove(save_dir + "/" + Name + str(count) + ".csv")
            writer = csv.writer(
                open(save_dir + "/" + Name + str(count) + ".csv", "a")
            )
            writer.writerows(Event_data)


def MovieLabelDataSet(
    image_dir,
    seg_image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Seg_path = os.path.join(seg_image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesSeg = glob.glob(Seg_path)
    filesSeg.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:

        name = os.path.basename(os.path.splitext(fname)[0])
        for Segfname in filesSeg:

            Segname = os.path.basename(os.path.splitext(Segfname)[0])

            if name == Segname:

                for csvfname in filesCsv:
                    count = 0
                    Csvname = os.path.basename(os.path.splitext(csvfname)[0])
                    for i in range(0, len(static_name)):
                        event_name = static_name[i]
                        trainlabel = static_label[i]
                        classfound = (
                            Csvname == csv_name_diff + event_name + name
                        )
                        if classfound:
                            print(Csvname)
                            image = imread(fname).astype(dtype)
                            segimage = imread(Segfname).astype("uint16")
                            dataset = pd.read_csv(csvfname)
                            time = dataset[dataset.keys()[0]][1:]
                            y = dataset[dataset.keys()[1]][1:]
                            x = dataset[dataset.keys()[2]][1:]

                            # Categories + XYHW + Confidence
                            for (key, t) in time.items():

                                MovieMaker(
                                    t,
                                    y[key],
                                    x[key],
                                    image,
                                    segimage,
                                    crop_size,
                                    gridx,
                                    gridy,
                                    total_categories,
                                    trainlabel,
                                    name + event_name + str(count),
                                    save_dir,
                                    tshift,
                                    normalizeimage,
                                    dtype,
                                )
                                count = count + 1
                            image = None
                            segimage = None


def VolumeLabelDataSet(
    image_dir,
    seg_image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    gridz=1,
    tshift=0,
    normalizeimage=True,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Seg_path = os.path.join(seg_image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesSeg = glob.glob(Seg_path)
    filesSeg.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for fname in files_raw:

        name = os.path.basename(os.path.splitext(fname)[0])
        for Segfname in filesSeg:

            Segname = os.path.basename(os.path.splitext(Segfname)[0])

            if name == Segname:

                for csvfname in filesCsv:
                    count = 0
                    Csvname = os.path.basename(os.path.splitext(csvfname)[0])
                    for i in range(0, len(static_name)):
                        event_name = static_name[i]
                        trainlabel = static_label[i]
                        classfound = (
                            Csvname == csv_name_diff + event_name + name
                        )
                        if classfound:
                            print(Csvname)
                            image = imread(fname).astype(dtype)
                            segimage = imread(Segfname).astype("uint16")
                            dataset = pd.read_csv(csvfname)
                            time = dataset[dataset.keys()[0]][1:]
                            z = dataset[dataset.keys()[1]][1:]
                            y = dataset[dataset.keys()[2]][1:]
                            x = dataset[dataset.keys()[3]][1:]

                            # Categories + XYZHW + Confidence
                            for (key, t) in time.items():

                                VolumeMaker(
                                    t,
                                    z[key],
                                    y[key],
                                    x[key],
                                    image,
                                    segimage,
                                    crop_size,
                                    gridx,
                                    gridy,
                                    gridz,
                                    total_categories,
                                    trainlabel,
                                    name + event_name + str(count),
                                    save_dir,
                                    tshift,
                                    normalizeimage,
                                    dtype,
                                )
                                count = count + 1
                            image = None
                            segimage = None


def VolumeMaker(
    time,
    z,
    y,
    x,
    image,
    segimage,
    crop_size,
    gridx,
    gridy,
    gridz,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
    normalizeimage,
    dtype,
):

    sizex, sizey, sizez, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy
    imagesizez = sizez * gridz
    shiftNone = [0, 0, 0]

    AllShifts = [shiftNone]

    time = time - tshift
    if normalizeimage:
        image = normalizeFloatZeroOne(
            image.astype(dtype), 1, 99.8, dtype=dtype
        )
    if time > size_tminus:

        # slice the images

        currentsegimage = segimage[int(time), :].astype("uint16")
        image_props = getHWD(
            x, y, z, currentsegimage, imagesizex, imagesizey, imagesizez
        )
        if image_props is not None:
            height, width, depth, center, seg_label = image_props
            smallimage = CreateVolume(
                image, size_tminus, size_tplus, int(time)
            )

            for shift in AllShifts:

                newname = name + "shift" + str(shift)
                Event_data = []
                newcenter = center
                x = center[2]
                y = center[1]
                z = center[0]

                Label = np.zeros([total_categories + 8])
                Label[trainlabel] = 1

                # T co ordinate
                Label[total_categories + 3] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                if (
                    x > sizex / 2
                    and z > sizez / 2
                    and y > sizey / 2
                    and z + int(imagesizez / 2) < image.shape[1]
                    and y + int(imagesizey / 2) < image.shape[2]
                    and x + int(imagesizex / 2) < image.shape[3]
                    and time > size_tminus
                    and time + size_tplus + 1 < image.shape[0]
                ):
                    crop_xminus = x - int(imagesizex / 2)
                    crop_xplus = x + int(imagesizex / 2)
                    crop_yminus = y - int(imagesizey / 2)
                    crop_yplus = y + int(imagesizey / 2)
                    crop_zminus = z - int(imagesizez / 2)
                    crop_zplus = z + int(imagesizez / 2)
                    region = (
                        slice(0, smallimage.shape[0]),
                        slice(int(crop_zminus), int(crop_zplus)),
                        slice(int(crop_yminus), int(crop_yplus)),
                        slice(int(crop_xminus), int(crop_xplus)),
                    )

                    # Define the movie region volume that was cut
                    crop_image = smallimage[region]
                    seglocationx = newcenter[2] - crop_xminus
                    seglocationy = newcenter[1] - crop_yminus
                    seglocationz = newcenter[0] - crop_zminus

                    Label[total_categories] = seglocationx / sizex
                    Label[total_categories + 1] = seglocationy / sizey
                    Label[total_categories + 2] = seglocationz / sizez
                    if height >= imagesizey:
                        height = 0.5 * imagesizey
                    if width >= imagesizex:
                        width = 0.5 * imagesizex
                    if depth >= imagesizez:
                        depth = 0.5 * imagesizez
                    # Height
                    Label[total_categories + 4] = height / imagesizey
                    # Width
                    Label[total_categories + 5] = width / imagesizex
                    # Depth
                    Label[total_categories + 6] = depth / imagesizez

                    Label[total_categories + 7] = 1
                    # Write the image as 32 bit tif file
                    if (
                        crop_image.shape[0] == size_tplus + size_tminus + 1
                        and crop_image.shape[1] == imagesizez
                        and crop_image.shape[2] == imagesizey
                        and crop_image.shape[3] == imagesizex
                    ):

                        imwrite(
                            (save_dir + "/" + newname + ".tif"),
                            crop_image.astype("float32"),
                        )
                        Event_data.append(
                            [Label[i] for i in range(0, len(Label))]
                        )
                        if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                            os.remove(save_dir + "/" + (newname) + ".csv")
                        writer = csv.writer(
                            open(save_dir + "/" + (newname) + ".csv", "a")
                        )
                        writer.writerows(Event_data)
    crop_image = None
    currentsegimage = None
    smallimage = None


def MovieMaker(
    time,
    y,
    x,
    image,
    segimage,
    crop_size,
    gridx,
    gridy,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
    normalizeimage,
    dtype,
):

    sizex, sizey, size_tminus, size_tplus = crop_size

    imagesizex = sizex * gridx
    imagesizey = sizey * gridy

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if normalizeimage:
        image = normalizeFloatZeroOne(
            image.astype(dtype), 1, 99.8, dtype=dtype
        )
    if time > size_tminus:
        currentsegimage = segimage[int(time), :].astype("uint16")
        image_props = getHW(x, y, currentsegimage, imagesizex, imagesizey)
        if image_props is not None:
            height, width, center, seg_label = image_props
            for shift in AllShifts:

                newname = name + "shift" + str(shift)
                Event_data = []
                newcenter = (center[0] - shift[1], center[1] - shift[0])
                x = center[1]
                y = center[0]

                Label = np.zeros([total_categories + 6])
                Label[trainlabel] = 1
                # T co ordinate
                Label[total_categories + 2] = (size_tminus) / (
                    size_tminus + size_tplus
                )
                smallimage = CreateVolume(
                    image, size_tminus, size_tplus, int(time)
                )

                if (
                    x + shift[0] > sizex / 2
                    and y + shift[1] > sizey / 2
                    and x + shift[0] + int(imagesizex / 2) < image.shape[2]
                    and y + shift[1] + int(imagesizey / 2) < image.shape[1]
                    and time > size_tminus
                    and time + size_tplus + 1 < image.shape[0]
                ):
                    crop_xminus = x - int(imagesizex / 2)
                    crop_xplus = x + int(imagesizex / 2)
                    crop_yminus = y - int(imagesizey / 2)
                    crop_yplus = y + int(imagesizey / 2)
                    region = (
                        slice(0, smallimage.shape[0]),
                        slice(
                            int(crop_yminus) + shift[1],
                            int(crop_yplus) + shift[1],
                        ),
                        slice(
                            int(crop_xminus) + shift[0],
                            int(crop_xplus) + shift[0],
                        ),
                    )
                    # Define the movie region volume that was cut
                    crop_image = smallimage[region]

                    seglocationx = newcenter[1] - crop_xminus
                    seglocationy = newcenter[0] - crop_yminus

                    Label[total_categories] = seglocationx / sizex
                    Label[total_categories + 1] = seglocationy / sizey
                    if height >= imagesizey:
                        height = 0.5 * imagesizey
                    if width >= imagesizex:
                        width = 0.5 * imagesizex
                    # Height
                    Label[total_categories + 3] = height / imagesizey
                    # Width
                    Label[total_categories + 4] = width / imagesizex
                    Label[total_categories + 5] = 1
                    # Write the image as 32 bit tif file
                    if (
                        crop_image.shape[0] == size_tplus + size_tminus + 1
                        and crop_image.shape[1] == imagesizey
                        and crop_image.shape[2] == imagesizex
                    ):

                        imwrite(
                            (save_dir + "/" + newname + ".tif"),
                            crop_image.astype("float32"),
                        )
                        Event_data.append(
                            [Label[i] for i in range(0, len(Label))]
                        )
                        if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                            os.remove(save_dir + "/" + (newname) + ".csv")
                        writer = csv.writer(
                            open(save_dir + "/" + (newname) + ".csv", "a")
                        )
                        writer.writerows(Event_data)
    image = None
    segimage = None
    crop_image = None
    currentsegimage = None
    smallimage = None


def Readname(fname):

    return os.path.basename(os.path.splitext(fname)[0])


def ImageLabelDataSet(
    image_dir,
    seg_image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    tshift=0,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Seg_path = os.path.join(seg_image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesSeg = glob.glob(Seg_path)
    filesSeg.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for csvfname in filesCsv:
        print(csvfname)
        count = 0
        Csvname = os.path.basename(os.path.splitext(csvfname)[0])

        for fname in files_raw:

            name = os.path.basename(os.path.splitext(fname)[0])
            for Segfname in filesSeg:

                Segname = os.path.basename(os.path.splitext(Segfname)[0])

                if name == Segname:

                    for i in range(0, len(static_name)):
                        event_name = static_name[i]
                        trainlabel = static_label[i]
                        if Csvname == csv_name_diff + name + event_name:
                            image = imread(fname).astype(dtype)
                            image = normalizeFloatZeroOne(
                                image.astype(dtype), 1, 99.8, dtype=dtype
                            )
                            segimage = imread(Segfname)
                            dataset = pd.read_csv(csvfname)
                            time = dataset[dataset.keys()[0]][1:]
                            y = dataset[dataset.keys()[1]][1:]
                            x = dataset[dataset.keys()[2]][1:]

                            # Categories + XYHW + Confidence
                            for (key, t) in time.items():
                                ImageMaker(
                                    t,
                                    y[key],
                                    x[key],
                                    image,
                                    segimage,
                                    crop_size,
                                    gridx,
                                    gridy,
                                    total_categories,
                                    trainlabel,
                                    name + event_name + str(count),
                                    save_dir,
                                    tshift,
                                )
                                count = count + 1


def SegFreeImageLabelDataSet(
    image_dir,
    csv_dir,
    save_dir,
    static_name,
    static_label,
    csv_name_diff,
    crop_size,
    gridx=1,
    gridy=1,
    dtype=np.uint8,
):

    raw_path = os.path.join(image_dir, "*tif")
    Csv_path = os.path.join(csv_dir, "*csv")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    Path(save_dir).mkdir(exist_ok=True)
    total_categories = len(static_name)

    for csvfname in filesCsv:
        print(csvfname)
        count = 0
        Csvname = os.path.basename(os.path.splitext(csvfname)[0])

        for fname in files_raw:

            name = os.path.basename(os.path.splitext(fname)[0])
            image = imread(fname).astype(dtype)

            image = normalizeFloatZeroOne(
                image.astype(dtype), 1, 99.8, dtype=dtype
            )
            for i in range(0, len(static_name)):
                event_name = static_name[i]
                trainlabel = static_label[i]
                if Csvname == csv_name_diff + name + event_name:
                    dataset = pd.read_csv(csvfname)
                    time = dataset[dataset.keys()[0]][1:]
                    y = dataset[dataset.keys()[1]][1:]
                    x = dataset[dataset.keys()[2]][1:]

                    # Categories + XYHW + Confidence
                    for (key, t) in time.items():
                        SegFreeImageMaker(
                            t,
                            y[key],
                            x[key],
                            image,
                            crop_size,
                            gridx,
                            gridy,
                            total_categories,
                            trainlabel,
                            name + event_name + str(count),
                            save_dir,
                        )
                        count = count + 1


def CreateVolume(patch, size_tminus, size_tplus, timepoint):
    starttime = timepoint - int(size_tminus)
    endtime = timepoint + int(size_tplus) + 1
    smallimg = patch[starttime:endtime, :]

    return smallimg


def createNPZ(
    save_dir,
    axes,
    save_name="oneat",
    save_name_val="oneatVal",
    expand=True,
    static=False,
    flip_channel_axis=False,
    train_size=0.95,
):

    data = []
    label = []

    raw_path = os.path.join(save_dir, "*tif")
    files_raw = glob.glob(raw_path)
    files_raw.sort
    NormalizeImages = [imread(fname) for fname in files_raw]

    names = [Readname(fname) for fname in files_raw]
    # Normalize everything before it goes inside the training
    for i in range(0, len(NormalizeImages)):

        n = NormalizeImages[i]

        blankX = n
        csvfname = save_dir + "/" + names[i] + ".csv"
        arr = []
        with open(csvfname) as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            arr = list(reader)[0]
            arr = np.array(arr)

        blankY = arr
        blankY = np.expand_dims(blankY, -1)
        if expand:

            blankX = np.expand_dims(blankX, -1)

        data.append(blankX)
        label.append(blankY)

    dataarr = np.asarray(data)
    labelarr = np.asarray(label)
    if flip_channel_axis:
        np.swapaxes(dataarr, 1, -1)
    if static:
        try:
            dataarr = dataarr[:, 0, :, :, :]
        except ValueError:
            pass
    print(dataarr.shape, labelarr.shape)
    traindata, validdata, trainlabel, validlabel = train_test_split(
        dataarr,
        labelarr,
        train_size=train_size,
        test_size=1 - train_size,
        shuffle=True,
    )
    save_full_training_data(save_dir, save_name, traindata, trainlabel, axes)
    save_full_training_data(
        save_dir, save_name_val, validdata, validlabel, axes
    )


def _raise(e):
    raise e


def ImageMaker(
    time,
    y,
    x,
    image,
    segimage,
    crop_size,
    gridX,
    gridY,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
):

    sizeX, sizeY = crop_size

    ImagesizeX = sizeX * gridX
    ImagesizeY = sizeY * gridY

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if time < segimage.shape[0] - 1 and time > 0:
        currentsegimage = segimage[int(time), :].astype("uint16")

        image_props = getHW(x, y, currentsegimage, ImagesizeX, ImagesizeY)
        if image_props is not None:
            height, width, center, seg_label = image_props
            for shift in AllShifts:

                newname = name + "shift" + str(shift)
                newcenter = (center[0] - shift[1], center[1] - shift[0])
                Event_data = []

                x = center[1]
                y = center[0]
                Label = np.zeros([total_categories + 5])
                Label[trainlabel] = 1
                if (
                    x + shift[0] > sizeX / 2
                    and y + shift[1] > sizeY / 2
                    and x + shift[0] + int(ImagesizeX / 2) < image.shape[2]
                    and y + shift[1] + int(ImagesizeY / 2) < image.shape[1]
                ):
                    crop_Xminus = x - int(ImagesizeX / 2)
                    crop_Xplus = x + int(ImagesizeX / 2)
                    crop_Yminus = y - int(ImagesizeY / 2)
                    crop_Yplus = y + int(ImagesizeY / 2)

                    for tex in range(int(time) - 2, int(time) + 2):
                        newname = newname + str(tex)
                        region = (
                            slice(int(tex - 1), int(tex)),
                            slice(
                                int(crop_Yminus) + shift[1],
                                int(crop_Yplus) + shift[1],
                            ),
                            slice(
                                int(crop_Xminus) + shift[0],
                                int(crop_Xplus) + shift[0],
                            ),
                        )

                        crop_image = image[region]

                        seglocationx = newcenter[1] - crop_Xminus
                        seglocationy = newcenter[0] - crop_Yminus

                        Label[total_categories] = seglocationx / sizeX
                        Label[total_categories + 1] = seglocationy / sizeY

                        if height >= ImagesizeY:
                            height = 0.5 * ImagesizeY
                        if width >= ImagesizeX:
                            width = 0.5 * ImagesizeX

                        Label[total_categories + 2] = height / ImagesizeY
                        Label[total_categories + 3] = width / ImagesizeX
                        Label[total_categories + 4] = 1

                        if (
                            crop_image.shape[1] == ImagesizeY
                            and crop_image.shape[2] == ImagesizeX
                        ):
                            imwrite(
                                (save_dir + "/" + newname + ".tif"),
                                crop_image.astype("float32"),
                            )
                            Event_data.append(
                                [Label[i] for i in range(0, len(Label))]
                            )
                            if os.path.exists(
                                save_dir + "/" + (newname) + ".csv"
                            ):
                                os.remove(save_dir + "/" + (newname) + ".csv")
                            writer = csv.writer(
                                open(save_dir + "/" + (newname) + ".csv", "a")
                            )
                            writer.writerows(Event_data)


def SegFreeImageMaker(
    time,
    y,
    x,
    image,
    crop_size,
    gridX,
    gridY,
    total_categories,
    trainlabel,
    name,
    save_dir,
    tshift,
):

    sizex, sizey = crop_size

    ImagesizeX = sizex * gridX
    ImagesizeY = sizey * gridY

    shiftNone = [0, 0]
    AllShifts = [shiftNone]

    time = time - tshift
    if time < image.shape[0] - 1 and time > 0:

        for shift in AllShifts:

            newname = name + "shift" + str(shift)
            newcenter = (y - shift[1], x - shift[0])
            Event_data = []

            Label = np.zeros([total_categories + 5])
            Label[trainlabel] = 1
            if (
                x + shift[0] > sizex / 2
                and y + shift[1] > sizey / 2
                and x + shift[0] + int(ImagesizeX / 2) < image.shape[2]
                and y + shift[1] + int(ImagesizeY / 2) < image.shape[1]
            ):
                crop_Xminus = x - int(ImagesizeX / 2)
                crop_Xplus = x + int(ImagesizeX / 2)
                crop_Yminus = y - int(ImagesizeY / 2)
                crop_Yplus = y + int(ImagesizeY / 2)

                for tex in range(int(time) - 2, int(time) + 2):
                    newname = newname + str(tex)
                    region = (
                        slice(int(tex - 1), int(tex)),
                        slice(
                            int(crop_Yminus) + shift[1],
                            int(crop_Yplus) + shift[1],
                        ),
                        slice(
                            int(crop_Xminus) + shift[0],
                            int(crop_Xplus) + shift[0],
                        ),
                    )

                    crop_image = image[region]

                    seglocationx = newcenter[1] - crop_Xminus
                    seglocationy = newcenter[0] - crop_Yminus

                    Label[total_categories] = seglocationx / sizex
                    Label[total_categories + 1] = seglocationy / sizey
                    Label[total_categories + 2] = 1
                    Label[total_categories + 3] = 1
                    Label[total_categories + 4] = 1

                    if (
                        crop_image.shape[1] == ImagesizeY
                        and crop_image.shape[2] == ImagesizeX
                    ):
                        imwrite(
                            (save_dir + "/" + newname + ".tif"),
                            crop_image.astype("float32"),
                        )
                        Event_data.append(
                            [Label[i] for i in range(0, len(Label))]
                        )
                        if os.path.exists(save_dir + "/" + (newname) + ".csv"):
                            os.remove(save_dir + "/" + (newname) + ".csv")
                        writer = csv.writer(
                            open(save_dir + "/" + (newname) + ".csv", "a")
                        )
                        writer.writerows(Event_data)


def getHW(defaultX, defaultY, currentsegimage, imagesizex, imagesizey):

    properties = measure.regionprops(currentsegimage)

    TwoDLocation = (defaultY, defaultX)
    if (
        int(TwoDLocation[0]) < currentsegimage.shape[0]
        and int(TwoDLocation[1]) < currentsegimage.shape[1]
    ):
        SegLabel = currentsegimage[int(TwoDLocation[0]), int(TwoDLocation[1])]
        for prop in properties:

            if SegLabel > 0 and prop.label == SegLabel:
                minr, minc, maxr, maxc = prop.bbox
                center = (defaultY, defaultX)
                height = abs(maxc - minc)
                width = abs(maxr - minr)
                return height, width, center, SegLabel

            if SegLabel == 0:

                center = (defaultY, defaultX)
                height = 0.5 * imagesizex
                width = 0.5 * imagesizey
                return height, width, center, SegLabel


def getHWD(
    defaultX,
    defaultY,
    defaultZ,
    currentsegimage,
    imagesizex,
    imagesizey,
    imagesizez,
):

    properties = measure.regionprops(currentsegimage)
    DLocation = (defaultZ, defaultY, defaultX)
    if (
        int(DLocation[0]) < currentsegimage.shape[0]
        and int(DLocation[1]) < currentsegimage.shape[1]
        and int(DLocation[2]) < currentsegimage.shape[2]
    ):
        SegLabel = currentsegimage[
            int(DLocation[0]), int(DLocation[1]), int(DLocation[2])
        ]
        for prop in properties:

            if SegLabel > 0 and prop.label == SegLabel:
                minr, minc, mind, maxr, maxc, maxd = prop.bbox
                center = (defaultZ, defaultY, defaultX)
                height = abs(maxc - minc)
                width = abs(maxr - minr)
                depth = abs(maxd - mind)
                return height, width, depth, center, SegLabel

            if SegLabel == 0:

                center = (defaultZ, defaultY, defaultX)
                height = 0.5 * imagesizex
                width = 0.5 * imagesizey
                depth = 0.5 * imagesizez
                return height, width, depth, center, SegLabel


def save_full_training_data(directory, filename, data, label, axes):
    """Save training data in ``.npz`` format."""

    len(axes) == data.ndim or _raise(ValueError())
    np.savez(directory + filename, data=data, label=label, axes=axes)


def InterchangeTXY(TXYCSV, save_dir):

    dataset = pd.read_csv(TXYCSV)
    time = dataset[dataset.keys()[0]][1:]
    x = dataset[dataset.keys()[1]][1:]
    y = dataset[dataset.keys()[2]][1:]

    Event_data = []

    Name = os.path.basename(os.path.splitext(TXYCSV)[0])

    for (key, t) in time.items():

        Event_data.append([t, y[key], x[key]])

    writer = csv.writer(open(save_dir + "/" + (Name) + ".csv", "a"))
    writer.writerows(Event_data)



#!/usr/bin/env python3
"""
Created on Wed Aug  4 14:50:47 2021

@author: vkapoor
"""
import csv
import glob
import json
import os
from pathlib import Path

import napari
import numpy as np
import pandas as pd
from qtpy.QtCore import Qt
from scipy import spatial
from tifffile import imread

from oneat.NEATUtils.oneat_animation._qt import OneatVolumeWidget, OneatWidget
from oneat.NEATUtils.oneat_animation.OneatVisualization import MidSlices


class NEATViz:
    def __init__(
        self,
        imagedir: str,
        csvdir: str,
        savedir: str,
        categories_json: dict,
        heatmapimagedir: str = None,
        segimagedir: str = None,
        heatname: str = "_Heat",
        eventname: str = "_Event",
        fileextension: str = "*tif",
        blur_radius: int = 5,
        start_project_mid: int = None,
        end_project_mid: int = None,
        headless: bool = False,
        volume: bool = False,
        batch: bool = False,
        event_threshold: float = 0.999,
        nms_space: int = 10,
        nms_time: int = 3,
    ):

        self.imagedir = imagedir
        self.heatmapimagedir = heatmapimagedir
        self.segimagedir = segimagedir
        self.savedir = savedir
        self.volume = volume
        self.csvdir = csvdir
        self.heatname = heatname
        self.eventname = eventname
        self.headless = headless
        self.batch = batch
        self.event_threshold = event_threshold
        self.nms_space = nms_space
        self.nms_time = nms_time
        self.categories_json = categories_json
        self.start_project_mid = start_project_mid
        self.end_project_mid = end_project_mid
        self.fileextension = fileextension
        self.blur_radius = blur_radius

        Path(self.savedir).mkdir(exist_ok=True)
        Path(self.csvdir).mkdir(exist_ok=True)
        self.viewer = napari.Viewer()

        self.time = 0
        self.key_categories = self.load_json()

        if not self.headless and not self.volume:
            self.showNapari()
        if self.headless and not self.volume:
            self.donotshowNapari()
        if self.volume:
            self.showVolumeNapari()

    def load_json(self):
        with open(self.categories_json) as f:
            return json.load(f)

    def donotshowNapari(self):

        Raw_path = os.path.join(self.imagedir, self.fileextension)
        X = glob.glob(Raw_path)

        for imagename in X:

            Name = os.path.basename(os.path.splitext(imagename)[0])
            image = imread(imagename)
            seg_image = imread(self.segimagedir + Name + ".tif")
            if (
                self.start_project_mid is not None
                or self.end_project_mid is not None
            ):
                if len(seg_image.shape) == 4:
                    seg_image = MidSlices(
                        seg_image,
                        self.start_project_mid,
                        self.end_project_mid,
                        False,
                        axis=1,
                    )
            headlesscall(
                image,
                imagename,
                self.key_categories,
                self.event_threshold,
                self.nms_space,
                self.nms_time,
                self.csvdir,
                self.savedir,
                self.batch,
            )

    def showNapari(self):

        self.oneat_widget = OneatWidget(
            self.viewer,
            self.csvdir,
            self.savedir,
            "Name",
            self.key_categories,
            segimagedir=self.segimagedir,
            heatimagedir=self.heatmapimagedir,
            heatname=self.heatname,
            start_project_mid=self.start_project_mid,
            end_project_mid=self.end_project_mid,
        )
        Raw_path = os.path.join(self.imagedir, self.fileextension)
        X = glob.glob(Raw_path)
        Imageids = []
        self.oneat_widget.frameWidget.imageidbox.addItem("Select Image")
        self.oneat_widget.frameWidget.eventidbox.addItem("Select Event")
        for imagename in X:
            Imageids.append(imagename)

        for i in range(0, len(Imageids)):
            self.oneat_widget.frameWidget.imageidbox.addItem(str(Imageids[i]))

        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                self.oneat_widget.frameWidget.eventidbox.addItem(event_name)

        dock_widget = self.viewer.window.add_dock_widget(
            self.oneat_widget, area="right"
        )
        self.viewer.window._qt_window.resizeDocks(
            [dock_widget], [200], Qt.Horizontal
        )

        napari.run()

    def showVolumeNapari(self):

        self.oneat_widget = OneatVolumeWidget(
            self.viewer,
            self.csvdir,
            self.savedir,
            "Name",
            self.key_categories,
            segimagedir=self.segimagedir,
            heatimagedir=self.heatmapimagedir,
            heatname=self.heatname,
        )
        Raw_path = os.path.join(self.imagedir, self.fileextension)
        X = glob.glob(Raw_path)
        Imageids = []
        self.oneat_widget.frameWidget.imageidbox.addItem("Select Image")
        self.oneat_widget.frameWidget.eventidbox.addItem("Select Event")
        for imagename in X:
            Imageids.append(imagename)

        for i in range(0, len(Imageids)):
            self.oneat_widget.frameWidget.imageidbox.addItem(str(Imageids[i]))

        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                self.oneat_widget.frameWidget.eventidbox.addItem(event_name)

        dock_widget = self.viewer.window.add_dock_widget(
            self.oneat_widget, area="right"
        )
        self.viewer.window._qt_window.resizeDocks(
            [dock_widget], [200], Qt.Horizontal
        )

        napari.run()


def cluster_points(
    event_locations_dict, event_locations_size_dict, nms_space, nms_time
):

    for (k, v) in event_locations_dict.items():
        currenttime = k
        event_locations = v

        tree = spatial.cKDTree(event_locations)
        for i in range(1, nms_time):

            forwardtime = currenttime + i
            if int(forwardtime) in event_locations_dict.keys():
                forward_event_locations = event_locations_dict[
                    int(forwardtime)
                ]
                for location in forward_event_locations:
                    if (
                        int(forwardtime),
                        int(location[0]),
                        int(location[1]),
                    ) in event_locations_size_dict:
                        forwardsize, forwardscore = event_locations_size_dict[
                            int(forwardtime),
                            int(location[0]),
                            int(location[1]),
                        ]
                        distance, nearest_location = tree.query(location)
                        nearest_location = int(
                            event_locations[nearest_location][0]
                        ), int(event_locations[nearest_location][1])

                        if distance <= nms_space:
                            if (
                                int(currenttime),
                                int(nearest_location[0]),
                                int(nearest_location[1]),
                            ) in event_locations_size_dict:
                                (
                                    currentsize,
                                    currentscore,
                                ) = event_locations_size_dict[
                                    int(currenttime),
                                    int(nearest_location[0]),
                                    int(nearest_location[1]),
                                ]
                                if currentsize >= forwardsize:
                                    event_locations_size_dict.pop(
                                        (
                                            int(forwardtime),
                                            int(location[0]),
                                            int(location[1]),
                                        )
                                    )

                                if currentsize < forwardsize:
                                    event_locations_size_dict.pop(
                                        (
                                            int(currenttime),
                                            int(nearest_location[0]),
                                            int(nearest_location[1]),
                                        )
                                    )
    return event_locations_size_dict


def headlesscall(
    image: np.ndarray,
    imagename: str,
    key_categories: dict,
    event_threshold: float,
    nms_space: int,
    nms_time: int,
    csvdir: str,
    savedir: str,
    batch: bool,
):
    for (event_name, event_label) in key_categories.items():
        if event_label > 0:
            event_locations = []
            size_locations = []
            score_locations = []
            event_locations = []
            confidence_locations = []
            event_locations_dict = {}
            event_locations_size_dict = {}
            if not batch:
                csvnames = [
                    csvdir
                    + "/"
                    + event_name
                    + "Location"
                    + (os.path.splitext(os.path.basename(imagename)))[0]
                    + ".csv"
                ]
                savename = (
                    event_name
                    + "Location"
                    + (os.path.splitext(os.path.basename(imagename)))[0]
                )
            else:
                csvnames = list(Path(csvdir).glob("*.csv"))
            for csvname in csvnames:
                event_locations = []
                size_locations = []
                score_locations = []
                event_locations = []
                confidence_locations = []
                event_locations_dict = {}
                event_locations_size_dict = {}
                savename = csvname.stem
                print(savename)
                dataset = pd.read_csv(csvname, delimiter=",")
                # Data is written as T, Y, X, Score, Size, Confidence
                T = dataset[dataset.keys()[0]][0:]
                Y = dataset[dataset.keys()[2]][0:]
                X = dataset[dataset.keys()[3]][0:]
                Score = dataset[dataset.keys()[4]][0:]
                Size = dataset[dataset.keys()[5]][0:]
                Confidence = dataset[dataset.keys()[6]][0:]
                listtime = T.tolist()
                listy = Y.tolist()
                listx = X.tolist()
                listsize = Size.tolist()

                listscore = Score.tolist()
                listconfidence = Confidence.tolist()

                for i in range(len(listtime)):

                    tcenter = int(listtime[i])
                    ycenter = listy[i]
                    xcenter = listx[i]
                    size = listsize[i]
                    score = listscore[i]
                    confidence = listconfidence[i]
                    if score > event_threshold:
                        event_locations.append(
                            [int(tcenter), int(ycenter), int(xcenter)]
                        )

                        if int(tcenter) in event_locations_dict.keys():
                            current_list = event_locations_dict[int(tcenter)]
                            current_list.append([int(ycenter), int(xcenter)])
                            event_locations_dict[int(tcenter)] = current_list
                            event_locations_size_dict[
                                (int(tcenter), int(ycenter), int(xcenter))
                            ] = [size, score]
                        else:
                            current_list = []
                            current_list.append([int(ycenter), int(xcenter)])
                            event_locations_dict[int(tcenter)] = current_list
                            event_locations_size_dict[
                                int(tcenter), int(ycenter), int(xcenter)
                            ] = [size, score]

                        size_locations.append(size)
                        score_locations.append(score)
                        confidence_locations.append(confidence)

                event_locations_size_dict = cluster_points(
                    event_locations_dict,
                    event_locations_size_dict,
                    nms_space,
                    nms_time,
                )
                event_locations_clean = []
                dict_locations = event_locations_size_dict.keys()
                tlocations = []
                zlocations = []
                ylocations = []
                xlocations = []
                scores = []
                radiuses = []
                confidences = []
                angles = []
                for location, sizescore in event_locations_size_dict.items():
                    tlocations.append(float(location[0]))
                    if len(image.shape) == 4:
                        zlocations.append(float(image.shape[1] // 2))
                    else:
                        zlocations.append(0)
                    ylocations.append(float(location[1]))
                    xlocations.append(float(location[2]))
                    scores.append(float(sizescore[1]))
                    radiuses.append(float(sizescore[0]))
                    confidences.append(1)
                    angles.append(2)
                for location in dict_locations:
                    event_locations_clean.append(location)

                event_count = np.column_stack(
                    [
                        tlocations,
                        zlocations,
                        ylocations,
                        xlocations,
                        scores,
                        radiuses,
                        confidences,
                        angles,
                    ]
                )
                event_count = sorted(
                    event_count, key=lambda x: x[0], reverse=False
                )

                event_data = []
                csvname = savedir + "/" + "Clean" + savename
                if os.path.exists(csvname + ".csv"):
                    os.remove(csvname + ".csv")
                writer = csv.writer(open(csvname + ".csv", "a", newline=""))
                filesize = os.stat(csvname + ".csv").st_size

                if filesize < 1:
                    writer.writerow(
                        [
                            "T",
                            "Z",
                            "Y",
                            "X",
                            "Score",
                            "Size",
                            "Confidence",
                            "Angle",
                        ]
                    )
                for line in event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []



import csv
import glob
import os
import random

import numpy as np


def IOU(ann, centroids):
    w, h = ann
    similarities = []

    for centroid in centroids:
        c_w, c_h = centroid

        if c_w >= w and c_h >= h:
            similarity = w * h / (c_w * c_h)
        elif c_w >= w and c_h <= h:
            similarity = w * c_h / (w * h + (c_w - w) * c_h)
        elif c_w <= w and c_h >= h:
            similarity = c_w * h / (w * h + c_w * (c_h - h))
        else:  # means both w,h are bigger than c_w and c_h respectively
            similarity = (c_w * c_h) / (w * h)
        similarities.append(similarity)  # will become (k,) shape

    return np.array(similarities)


def avg_IOU(anns, centroids):
    n, d = anns.shape
    sum = 0.0

    for i in range(anns.shape[0]):
        sum += max(IOU(anns[i], centroids))

    return sum / n


def print_anchors(centroids):
    anchors = centroids.copy()

    widths = anchors[:, 0]
    sorted_indices = np.argsort(widths)

    r = "anchors: ["
    for i in sorted_indices[:-1]:
        r += f"{anchors[i, 0]:0.2f},{anchors[i, 1]:0.2f}, "

    # there should not be comma after last anchor, that's why
    r += "{:0.2f},{:0.2f}".format(
        anchors[sorted_indices[-1:], 0], anchors[sorted_indices[-1:], 1]
    )
    r += "]"

    print(r)


def run_kmeans(ann_dims, anchor_num):
    ann_num = ann_dims.shape[0]
    prev_assignments = np.ones(ann_num) * (-1)
    iteration = 0
    old_distances = np.zeros((ann_num, anchor_num))

    indices = [random.randrange(ann_dims.shape[0]) for i in range(anchor_num)]
    centroids = ann_dims[indices]
    anchor_dim = ann_dims.shape[1]

    while True:
        distances = []
        iteration += 1
        for i in range(ann_num):
            d = 1 - IOU(ann_dims[i], centroids)
            distances.append(d)
        distances = np.array(
            distances
        )  # distances.shape = (ann_num, anchor_num)

        print(
            f"iteration {iteration}: dists = {np.sum(np.abs(old_distances-distances))}"
        )

        # assign samples to centroids
        assignments = np.argmin(distances, axis=1)

        if (assignments == prev_assignments).all():
            return centroids

        # calculate new centroids
        centroid_sums = np.zeros((anchor_num, anchor_dim), np.float)
        for i in range(ann_num):
            centroid_sums[assignments[i]] += ann_dims[i]
        for j in range(anchor_num):
            centroids[j] = centroid_sums[j] / (np.sum(assignments == j) + 1e-6)

        prev_assignments = assignments.copy()
        old_distances = distances.copy()


def main():
    CSVDir = "/data/u934/service_imagerie/v_kapoor/FinalONEATTraining/Bin2OYoloneatV1/"
    Csv_path = os.path.join(CSVDir, "*csv")
    filesCsv = glob.glob(Csv_path)
    filesCsv.sort
    num_anchors = 5

    # run k_mean to find the anchors
    annotation_dims = []
    for csvfname in filesCsv:

        with open(csvfname) as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            for train_vec in reader:

                arr = [float(s) for s in train_vec[0:]]
        train_vec = arr

        relative_w = float(train_vec[6])
        relative_h = float(train_vec[7])

        annotation_dims.append(tuple(map(float, (relative_w, relative_h))))

    annotation_dims = np.array(annotation_dims)
    centroids = run_kmeans(annotation_dims, num_anchors)

    # write anchors to file
    print(
        "\naverage IOU for",
        num_anchors,
        "anchors:",
        "%0.2f" % avg_IOU(annotation_dims, centroids),
    )
    print_anchors(centroids)


if __name__ == "__main__":

    main()



import csv
import os
from pathlib import Path

import numpy as np
import pandas as pd
import tensorflow as tf
from keras import callbacks, optimizers
from keras.models import load_model
from scipy.optimize import curve_fit
from tqdm import tqdm

from oneat.NEATModels import nets
from oneat.NEATModels.loss import dynamic_yolo_loss
from oneat.NEATModels.nets import Concat
from oneat.NEATUtils import plotters, utils
from oneat.NEATUtils.utils import (
    focyoloprediction,
    load_json,
    normalizeFloatZeroOne,
    simpleaveragenms,
)

Boxname = "ImageIDBox"
EventBoxname = "EventIDBox"


class NEATFocus:

    """
    Parameters
    ----------

    NpzDirectory : Specify the location of npz file containing the training data with movies and labels

    TrainModelName : Specify the name of the npz file containing training data and labels

    ValidationModelName :  Specify the name of the npz file containing validation data and labels

    categories : Number of action classes

    Categories_Name : List of class names and labels

    model_dir : Directory location where trained model weights are to be read or written from


    model_keras : The model as it appears as a Keras function

    model_weights : If re-training model_weights = model_dir  else None as default


    epochs :  Number of training epochs, 55 by default

    batch_size : batch_size to be used for training, 20 by default



    """

    def __init__(self, config, model_dir, catconfig=None, cordconfig=None):

        self.config = config
        self.catconfig = catconfig
        self.cordconfig = cordconfig
        self.model_dir = model_dir
        if self.config is not None:
            self.npz_directory = config.npz_directory
            self.npz_name = config.npz_name
            self.npz_val_name = config.npz_val_name
            self.key_categories = config.key_categories
            self.stage_number = config.stage_number
            self.last_conv_factor = config.last_conv_factor
            self.show = config.show
            self.key_cord = config.key_cord
            self.box_vector = len(config.key_cord)
            self.categories = len(config.key_categories)
            self.depth = config.depth
            self.start_kernel = config.start_kernel
            self.mid_kernel = config.mid_kernel
            self.learning_rate = config.learning_rate
            self.epochs = config.epochs
            self.startfilter = config.startfilter
            self.batch_size = config.batch_size
            self.multievent = config.multievent
            self.imagex = config.imagex
            self.imagey = config.imagey
            self.imagez = config.size_tminus + config.size_tplus + 1
            self.size_zminus = config.size_tminus
            self.size_zplus = config.size_tplus
            self.nboxes = 1
            self.gridx = 1
            self.gridy = 1
            self.gridz = 1
            self.yolo_v0 = True
            self.yolo_v1 = False
            self.yolo_v2 = False
            self.stride = config.last_conv_factor
        if self.config is None:

            self.config = load_json(
                os.path.join(self.model_dir, "parameters.json")
            )
            self.npz_directory = self.config["npz_directory"]
            self.npz_name = self.config["npz_name"]
            self.npz_val_name = self.config["npz_val_name"]
            self.key_categories = self.catconfig
            self.box_vector = self.config["box_vector"]
            self.show = self.config["show"]
            self.key_cord = self.cordconfig
            self.categories = len(self.catconfig)
            self.depth = self.config["depth"]
            self.start_kernel = self.config["start_kernel"]
            self.mid_kernel = self.config["mid_kernel"]
            self.learning_rate = self.config["learning_rate"]
            self.epochs = self.config["epochs"]
            self.startfilter = self.config["startfilter"]
            self.batch_size = self.config["batch_size"]
            self.multievent = self.config["multievent"]
            self.imagex = self.config["imagex"]
            self.imagey = self.config["imagey"]
            self.imagez = (
                self.config["size_tminus"] + self.config["size_tplus"] + 1
            )
            self.size_zminus = self.config["size_tminus"]
            self.size_zplus = self.config["size_tplus"]
            self.nboxes = 1
            self.stage_number = self.config["stage_number"]
            self.last_conv_factor = self.config["last_conv_factor"]
            self.gridx = 1
            self.gridy = 1
            self.gridz = 1
            self.yolo_v0 = False
            self.yolo_v1 = True
            self.yolo_v2 = False
            self.stride = self.config["last_conv_factor"]

        self.X = None
        self.Y = None
        self.axes = None
        self.X_val = None
        self.Y_val = None
        self.Trainingmodel = None
        self.Xoriginal = None
        self.Xoriginal_val = None

        self.model_keras = nets.VollNet
        if self.multievent:
            self.last_activation = "sigmoid"
            self.entropy = "binary"

        if not self.multievent:
            self.last_activation = "softmax"
            self.entropy = "notbinary"
        self.yolo_loss = dynamic_yolo_loss(
            self.categories,
            self.gridx,
            self.gridy,
            self.gridz,
            1,
            self.box_vector,
            self.entropy,
            self.yolo_v0,
            self.yolo_v1,
            self.yolo_v2,
        )

    def loadData(self):

        (X, Y), axes = utils.load_full_training_data(
            self.npz_directory, self.npz_name, verbose=True
        )

        (X_val, Y_val), axes = utils.load_full_training_data(
            self.npz_directory, self.npz_val_name, verbose=True
        )

        self.Xoriginal = X
        self.Xoriginal_val = X_val

        self.X = X
        self.Y = Y[:, :, 0]
        self.X_val = X_val
        self.Y_val = Y_val[:, :, 0]

        self.axes = axes
        self.Y = self.Y.reshape((self.Y.shape[0], 1, 1, 1, self.Y.shape[1]))
        self.Y_val = self.Y_val.reshape(
            (self.Y_val.shape[0], 1, 1, 1, self.Y_val.shape[1])
        )

    def TrainModel(self):

        input_shape = (
            self.X.shape[1],
            self.X.shape[2],
            self.X.shape[3],
            self.X.shape[4],
        )

        print(input_shape)
        print(self.last_activation)
        Path(self.model_dir).mkdir(exist_ok=True)

        model_weights = os.path.join(self.model_dir, "weights.h5")
        if os.path.exists(model_weights):

            self.model_weights = model_weights
            print("loading weights")
        else:

            self.model_weights = None

        dummyY = np.zeros(
            [
                self.Y.shape[0],
                1,
                self.Y.shape[1],
                self.Y.shape[2],
                self.categories + self.box_vector,
            ]
        )
        dummyY[:, :, :, :, : self.Y.shape[4]] = self.Y

        dummyY_val = np.zeros(
            [
                self.Y_val.shape[0],
                1,
                self.Y_val.shape[1],
                self.Y_val.shape[2],
                self.categories + self.box_vector,
            ]
        )
        dummyY_val[:, :, :, :, : self.Y_val.shape[4]] = self.Y_val

        self.Y = dummyY
        self.Y_val = dummyY_val

        self.Trainingmodel = self.model_keras(
            input_shape,
            self.categories,
            box_vector=self.box_vector,
            stage_number=self.stage_number,
            last_conv_factor=self.last_conv_factor,
            depth=self.depth,
            start_kernel=self.start_kernel,
            mid_kernel=self.mid_kernel,
            startfilter=self.startfilter,
            input_weights=self.model_weights,
            last_activation=self.last_activation,
        )

        sgd = optimizers.SGD(
            lr=self.learning_rate, momentum=0.99, decay=1e-6, nesterov=True
        )
        self.Trainingmodel.compile(
            optimizer=sgd, loss=self.yolo_loss, metrics=["accuracy"]
        )

        self.Trainingmodel.summary()

        # Keras callbacks
        lrate = callbacks.ReduceLROnPlateau(
            monitor="loss", factor=0.1, patience=4, verbose=1
        )
        hrate = callbacks.History()
        srate = callbacks.ModelCheckpoint(
            self.model_dir,
            monitor="loss",
            verbose=1,
            save_best_only=False,
            save_weights_only=False,
            mode="auto",
            period=1,
        )
        prate = plotters.PlotHistory(
            self.Trainingmodel,
            self.X_val,
            self.Y_val,
            self.key_categories,
            self.key_cord,
            self.gridx,
            self.gridy,
            plot=self.show,
            nboxes=1,
        )

        # Train the model and save as a h5 file
        self.Trainingmodel.fit(
            self.X,
            self.Y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            validation_data=(self.X_val, self.Y_val),
            shuffle=True,
            callbacks=[lrate, hrate, srate, prate],
        )

        self.Trainingmodel.save(model_weights)

    def predict(
        self,
        image: np.ndarray,
        savedir: str = None,
        n_tiles: tuple = (1, 1),
        overlap_percent: float = 0.8,
        event_threshold: float = 0.5,
        iou_threshold: float = 0.0001,
        radius: int = 10,
        normalize: bool = True,
        activations: bool = False,
    ):

        self.image = image
        self.Colorimage = np.zeros(
            [self.image.shape[0], self.image.shape[1], self.image.shape[2], 3],
            dtype="uint16",
        )
        self.Maskimage = np.zeros(
            [self.image.shape[0], self.image.shape[1], self.image.shape[2], 3],
            dtype="float32",
        )
        self.image_mask_c1 = np.zeros(self.image.shape, dtype="float32")
        self.image_mask_c2 = np.zeros(self.image.shape, dtype="float32")
        self.Colorimage[:, :, :, 0] = self.image
        self.Maskimage[:, :, :, 0] = self.image
        self.radius = radius
        self.savedir = savedir
        self.n_tiles = n_tiles
        self.overlap_percent = overlap_percent
        self.iou_threshold = iou_threshold
        self.event_threshold = event_threshold
        self.normalize = normalize
        self.activations = activations
        self.iou_classedboxes = {}
        self.all_iou_classedboxes = {}
        self.maskboxes = {}

        self.model = self._build()

        self.first_pass_predict()

    def _build(self):

        model_weights = os.path.join(self.model_dir, "weights.h5")
        Model = load_model(
            model_weights,
            custom_objects={"loss": self.yolo_loss, "Concat": Concat},
        )

        return Model

    def first_pass_predict(self):

        eventboxes = []
        classedboxes = {}
        print("Detecting focus planes")
        if self.normalize:
            self.image = normalizeFloatZeroOne(self.image, 1, 99.8)

        for inputz in tqdm(range(0, self.image.shape[0])):
            if inputz <= self.image.shape[0] - self.imagez:

                eventboxes = []
                self.currentZ = inputz
                smallimage = CreateVolume(self.image, self.imagez, inputz)

                # self.current_Zpoints = [(j,k) for j in range(smallimage.shape[1]) for k in range(smallimage.shape[2]) ]
                # Cut off the region for training movie creation
                # Break image into tiles if neccessary
                predictions, allx, ally = self.predict_main(smallimage)
                # Iterate over tiles
                for p in range(0, len(predictions)):

                    sum_z_prediction = predictions[p]

                    if sum_z_prediction is not None:
                        # For each tile the prediction vector has shape N H W Categories + Training Vector labels
                        for i in range(0, sum_z_prediction.shape[0]):
                            z_prediction = sum_z_prediction[i]
                            boxprediction = focyoloprediction(
                                ally[p],
                                allx[p],
                                z_prediction,
                                self.stride,
                                inputz,
                                self.config,
                                self.key_categories,
                            )

                            if boxprediction is not None:
                                eventboxes = eventboxes + boxprediction

                for (event_name, event_label) in self.key_categories.items():

                    if event_label > 0:
                        current_event_box = []
                        for box in eventboxes:

                            event_prob = box[event_name]
                            if event_prob >= self.event_threshold:

                                current_event_box.append(box)
                        classedboxes[event_name] = [current_event_box]

                self.classedboxes = classedboxes
                self.eventboxes = eventboxes
                self.nms()
                if self.savedir is not None:
                    self.to_csv()
                self.draw()
                eventboxes = []
                classedboxes = {}

        self.print_planes()
        self.fit_curve()
        self.genmap()

    def nms(self):

        best_iou_classedboxes = {}
        all_best_iou_classedboxes = {}
        self.all_iou_classedboxes = {}
        self.iou_classedboxes = {}
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                # Get all events

                sorted_event_box = self.classedboxes[event_name][0]

                sorted_event_box = sorted(
                    sorted_event_box, key=lambda x: x[event_name], reverse=True
                )

                scores = [
                    sorted_event_box[i][event_name]
                    for i in range(len(sorted_event_box))
                ]
                best_sorted_event_box, all_boxes = simpleaveragenms(
                    sorted_event_box,
                    scores,
                    self.iou_threshold,
                    self.event_threshold,
                    event_name,
                )
                all_best_iou_classedboxes[event_name] = [all_boxes]
                best_iou_classedboxes[event_name] = [best_sorted_event_box]
                if self.activations:
                    if event_name in self.all_iou_classedboxes:
                        boxes = self.all_iou_classedboxes[event_name]
                        boxes.append(best_sorted_event_box)
                        self.all_iou_classedboxes[event_name] = boxes
                    else:
                        self.all_iou_classedboxes[
                            event_name
                        ] = best_sorted_event_box
        self.iou_classedboxes = best_iou_classedboxes
        self.all_iou_classedboxes = all_best_iou_classedboxes

    def to_csv(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:
                zlocations = []
                scores = []
                max_scores = []
                iou_current_event_box = self.iou_classedboxes[event_name][0]
                zcenter = iou_current_event_box["real_z_event"]
                max_score = iou_current_event_box["max_score"]
                score = iou_current_event_box[event_name]

                zlocations.append(zcenter)
                scores.append(score)
                max_scores.append(max_score)
                print(zlocations, scores)
                event_count = np.column_stack([zlocations, scores, max_scores])
                event_count = sorted(
                    event_count, key=lambda x: x[0], reverse=False
                )
                event_data = []
                csvname = self.savedir + "/" + event_name + "_FocusQuality"
                writer = csv.writer(open(csvname + ".csv", "a"))
                filesize = os.stat(csvname + ".csv").st_size
                if filesize < 1:
                    writer.writerow(["Z", "Score", "Max_score"])
                for line in event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []

    def fit_curve(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:
                readcsvname = self.savedir + "/" + event_name + "_FocusQuality"
                self.dataset = pd.read_csv(readcsvname, delimiter=",")
                self.dataset_index = self.dataset.index

                Z = self.dataset[self.dataset.keys()[0]][1:]
                score = self.dataset[self.dataset.keys()[1]][1:]

                H, A, mu0, sigma = gauss_fit(np.array(Z), np.array(score))
                csvname = (
                    self.savedir + "/" + event_name + "_GaussFitFocusQuality"
                )
                writer = csv.writer(open(csvname + ".csv", "a"))
                filesize = os.stat(csvname + ".csv").st_size
                if filesize < 1:
                    writer.writerow(["Amplitude", "Mean", "Sigma"])
                    writer.writerow([A, mu0, sigma])

    def print_planes(self):
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                csvfname = (
                    self.savedir + "/" + event_name + "_focus_quality" + ".csv"
                )
                dataset = pd.read_csv(csvfname, skiprows=0)
                z = dataset[dataset.keys()[0]][1:]
                score = dataset[dataset.keys()[1]][1:]
                maxz = z[np.argmax(score)] + 2
                print("Best Zs" + "for" + event_name + "at" + str(maxz))

    def draw(self):

        for (event_name, event_label) in self.key_categories.items():

            if event_label > 0:

                xlocations = []
                ylocations = []
                scores = []
                zlocations = []
                heights = []
                widths = []
                iou_current_event_boxes = self.all_iou_classedboxes[
                    event_name
                ][0]

                for iou_current_event_box in iou_current_event_boxes:

                    xcenter = iou_current_event_box["xcenter"]
                    ycenter = iou_current_event_box["ycenter"]
                    zcenter = iou_current_event_box["real_z_event"]
                    xstart = iou_current_event_box["xstart"]
                    ystart = iou_current_event_box["ystart"]
                    xend = xstart + iou_current_event_box["width"]
                    yend = ystart + iou_current_event_box["height"]
                    score = iou_current_event_box[event_name]

                    if event_label == 1:
                        for x in range(int(xstart), int(xend)):
                            for y in range(int(ystart), int(yend)):
                                if (
                                    y < self.image.shape[1]
                                    and x < self.image.shape[2]
                                ):
                                    self.Maskimage[int(zcenter), y, x, 1] = (
                                        self.Maskimage[int(zcenter), y, x, 1]
                                        + score
                                    )
                    else:

                        for x in range(int(xstart), int(xend)):
                            for y in range(int(ystart), int(yend)):
                                if (
                                    y < self.image.shape[1]
                                    and x < self.image.shape[2]
                                ):
                                    self.Maskimage[int(zcenter), y, x, 2] = (
                                        self.Maskimage[int(zcenter), y, x, 2]
                                        + score
                                    )

                    if score > 0.9:

                        xlocations.append(round(xcenter))
                        ylocations.append(round(ycenter))
                        scores.append(score)
                        zlocations.append(zcenter)
                        heights.append(iou_current_event_box["height"])
                        widths.append(iou_current_event_box["width"])

    def overlaptiles(self, sliceregion):

        if self.n_tiles == (1, 1):
            patch = []
            rowout = []
            column = []
            patchx = sliceregion.shape[2] // self.n_tiles[0]
            patchy = sliceregion.shape[1] // self.n_tiles[1]
            patchshape = (patchy, patchx)
            smallpatch, smallrowout, smallcolumn = chunk_list(
                sliceregion, patchshape, self.stride, [0, 0]
            )
            patch.append(smallpatch)
            rowout.append(smallrowout)
            column.append(smallcolumn)

        else:
            patchx = sliceregion.shape[2] // self.n_tiles[0]
            patchy = sliceregion.shape[1] // self.n_tiles[1]

            if patchx > self.imagex and patchy > self.imagey:
                if self.overlap_percent > 1 or self.overlap_percent < 0:
                    self.overlap_percent = 0.8

                jumpx = int(self.overlap_percent * patchx)
                jumpy = int(self.overlap_percent * patchy)

                patchshape = (patchy, patchx)
                rowstart = 0
                colstart = 0
                pairs = []
                # row is y, col is x

                while rowstart < sliceregion.shape[1] - patchy:
                    colstart = 0
                    while colstart < sliceregion.shape[2] - patchx:

                        # Start iterating over the tile with jumps = stride of the fully convolutional network.
                        pairs.append([rowstart, colstart])
                        colstart += jumpx
                    rowstart += jumpy

                # Include the last patch
                rowstart = sliceregion.shape[1] - patchy
                colstart = 0
                while colstart < sliceregion.shape[2]:
                    pairs.append([rowstart, colstart])
                    colstart += jumpx
                rowstart = 0
                colstart = sliceregion.shape[2] - patchx
                while rowstart < sliceregion.shape[1]:
                    pairs.append([rowstart, colstart])
                    rowstart += jumpy

                if (
                    sliceregion.shape[1] >= self.imagey
                    and sliceregion.shape[2] >= self.imagex
                ):

                    patch = []
                    rowout = []
                    column = []
                    for pair in pairs:
                        smallpatch, smallrowout, smallcolumn = chunk_list(
                            sliceregion, patchshape, self.stride, pair
                        )
                        if (
                            smallpatch.shape[1] >= self.imagey
                            and smallpatch.shape[2] >= self.imagex
                        ):
                            patch.append(smallpatch)
                            rowout.append(smallrowout)
                            column.append(smallcolumn)

            else:

                patch = []
                rowout = []
                column = []
                patchx = sliceregion.shape[2] // self.n_tiles[0]
                patchy = sliceregion.shape[1] // self.n_tiles[1]
                patchshape = (patchy, patchx)
                smallpatch, smallrowout, smallcolumn = chunk_list(
                    sliceregion, patchshape, self.stride, [0, 0]
                )
                patch.append(smallpatch)
                rowout.append(smallrowout)
                column.append(smallcolumn)
        self.patch = patch
        self.sy = rowout
        self.sx = column

    def predict_main(self, sliceregion):
        try:
            self.overlaptiles(sliceregion)
            predictions = []
            allx = []
            ally = []
            if len(self.patch) > 0:
                for i in range(0, len(self.patch)):

                    sum_time_prediction = self.make_patches(self.patch[i])
                    predictions.append(sum_time_prediction)
                    allx.append(self.sx[i])
                    ally.append(self.sy[i])

            else:

                sum_time_prediction = self.make_patches(self.patch)
                predictions.append(sum_time_prediction)
                allx.append(self.sx)
                ally.append(self.sy)

        except tf.errors.ResourceExhaustedError:

            print("Out of memory, increasing overlapping tiles for prediction")

            self.list_n_tiles = list(self.n_tiles)
            self.list_n_tiles[0] = self.n_tiles[0] + 1
            self.list_n_tiles[1] = self.n_tiles[1] + 1
            self.n_tiles = tuple(self.list_n_tiles)

            self.predict_main(sliceregion)

        return predictions, allx, ally

    def make_patches(self, sliceregion):

        predict_im = np.expand_dims(sliceregion, 0)

        prediction_vector = self.model.predict(
            np.expand_dims(predict_im, -1), verbose=0
        )

        return prediction_vector

    def make_batch_patches(self, sliceregion):

        prediction_vector = self.model.predict(
            np.expand_dims(sliceregion, -1), verbose=0
        )
        return prediction_vector


def chunk_list(image, patchshape, stride, pair):
    rowstart = pair[0]
    colstart = pair[1]

    endrow = rowstart + patchshape[0]
    endcol = colstart + patchshape[1]

    if endrow > image.shape[1]:
        endrow = image.shape[1]
    if endcol > image.shape[2]:
        endcol = image.shape[2]

    region = (
        slice(0, image.shape[0]),
        slice(rowstart, endrow),
        slice(colstart, endcol),
    )

    # The actual pixels in that region.
    patch = image[region]

    # Always normalize patch that goes into the netowrk for getting a prediction score

    return patch, rowstart, colstart


def CreateVolume(patch, imagez, timepoint):

    starttime = timepoint
    endtime = timepoint + imagez
    smallimg = patch[starttime:endtime, :]

    return smallimg


def normalizeZeroOne(x):
    x = x.astype("float32")

    minVal = np.min(x)
    maxVal = np.max(x)

    x = (x - minVal) / (maxVal - minVal + 1.0e-20)

    return x


def gauss(x, H, A, x0, sigma):
    return H + A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))


def gauss_fit(x, y):
    mean = sum(x * y) / sum(y)
    sigma = np.sqrt(sum(y * (x - mean) ** 2) / sum(y))
    popt, pcov = curve_fit(gauss, x, y, p0=[min(y), max(y), mean, sigma])
    return popt
